#include <vo_stdio.h>
#include <volink.h>     // Linker directives like DLLENTRY
#include <apploader.h>  // RunLibraryFunction etc
#include <string.h>
//#include <mutex.h>
//#include <kernel.h>
#include <stdlib.h>
#include <timers.h>
#include <audio.h>
#include "time.h"
//#include "rtc.h"

#include "vs1010c.h"
#include <vs1010dRom.h>

void StartAndWaitRtcReady(register u_int16 rtcCfBits) {
  PERIP(RTC_CF) = rtcCfBits;
  while (PERIP(RTC_CF));
}
 #if 0

u_int16 WriteRtcMem(register u_int16 addr, register u_int16 data) {
  PERIP(RTC_LOW) = 0;			/* Clear LSbs */
  PERIP(RTC_HIGH) = RTC_I_MEM_WR;	/* Set memory write instruction */
  StartAndWaitRtcReady(RTC_CF_IBUSY);	/* Wait for instruction ready */
  PERIP(RTC_LOW) = addr;		/* Set memory address */
  PERIP(RTC_HIGH) = data;		/* Set data */
  StartAndWaitRtcReady(RTC_CF_DBUSY);	/* Wait for data write to register */
  PERIP(RTC_CF) = RTC_CF_GSCK|RTC_CF_EXEC;
  while(PERIP(RTC_CF) & RTC_CF_GSCK);	/* Wait for data write to memory */
  PERIP(RTC_CF) = 0;			/* Clear execute */
}

u_int16 ReadRtcMem(register u_int16 addr) {
  PERIP(RTC_HIGH) = RTC_I_MEM_RD;	/* Set memory read instruction */
  StartAndWaitRtcReady(RTC_CF_IBUSY);	/* Wait for instruction ready */
  PERIP(RTC_LOW) = addr;		/* Set memory address */
  StartAndWaitRtcReady(RTC_CF_DBUSY);	/* Wait for address transmit ready */
  StartAndWaitRtcReady(RTC_CF_RDBUSY);	/* Wait for result */
  return PERIP(RTC_HIGH);
}
#endif

#define RTC_MAGIC_ADDRESS  0x0000
#define RTC_MAGIC_DATA     0x1e0b

/* RTC cannot be found by the hardware; probably not powered */
#define RTC_NOT_FOUND	0xFFFFFFFEUL
/* RTC can be found but time has not been set */
#define RTC_NOT_SET	0xFFFFFFFFUL


void WriteShifter(u_int32 data) {

	PERIP(RTC_HIGH) = 0;
 	StartAndWaitRtcReady(RTC_CF_IBUSY);

	PERIP(RTC_HIGH) = (u_int16)(data >> 16);
	PERIP(RTC_LOW) = (u_int16)data;
	StartAndWaitRtcReady(RTC_CF_DBUSY);
}

u_int32 ReadShifter() {

	u_int32 shifter;

	StartAndWaitRtcReady(RTC_CF_RDBUSY);
	
	shifter = (((u_int32)PERIP(RTC_HIGH)) << 16) | (u_int32)PERIP(RTC_LOW); 
	WriteShifter(shifter);
	
	return shifter;
}


unsigned long GetRtc(void) {

	u_int32 shifter = ReadShifter();
	u_int32 rtc;
	
	PERIP(RTC_HIGH) = RTC_I_READRTC;
	StartAndWaitRtcReady(RTC_CF_IBUSY);
 	
	StartAndWaitRtcReady(RTC_CF_RDBUSY);
  	rtc = (((u_int32)PERIP(RTC_HIGH)) << 16) | (u_int32)PERIP(RTC_LOW);

	WriteShifter(shifter);
  	
  	return rtc;	
}

u_int32 SetRtc(register u_int32 t) {

	u_int32 shifter = ReadShifter();

  	PERIP(RTC_HIGH) = RTC_I_LOADRTC;
	StartAndWaitRtcReady(RTC_CF_IBUSY);
	
	PERIP(RTC_HIGH) = (u_int16)(t >> 16);
	PERIP(RTC_LOW) = (u_int16)t;
	StartAndWaitRtcReady(RTC_CF_DBUSY);
 
	PERIP(RTC_CF) = RTC_CF_EXEC;
	printf("Delay starts \n");
	DelayL(10000000);
	printf("Delay ends \n");
	PERIP(RTC_CF) = 0;

	WriteShifter(shifter);

	return t;
}


ioresult SetDate(register char *p, register s_int16 nParam) {
  int set = 0;
  int i;
  time_t tim = time(NULL);
  struct tm *tm = localtime(&tim);
  ioresult res = S_ERROR;

  while (nParam--) {
    char *setMon=NULL;
    int setTim=0;
    int len = strlen(p);
    if (len == 10 && p[4] == '-' && p[7] == '-')  {
      tm->tm_year = -1900;
      setMon = p+5;
    } else if (len == 8 && p[2] == '-' && p[5] == '-')  {
      tm->tm_year = 100;
      setMon = p+3;
    } else if (len == 8 && p[2] == ':' && p[5] == ':')  {
      tm->tm_sec = atoi(p+6);
      setTim = 1;
    } else if (len == 5 && p[2] == ':')  {
      tm->tm_sec = 0;
      setTim = 1;
    } else {
      printf("E: Malformed parameter \"%s\"\n", p);
      goto finally;
    }

    if (setMon) {
      tm->tm_year += atoi(p+0);
      tm->tm_mon = atoi(setMon)-1;
      tm->tm_mday = atoi(setMon+3);
      set |= 1;
    }

    if (setTim) {
      tm->tm_hour = atoi(p+0);
      tm->tm_min = atoi(p+3);
      set |= 2;
    }

    p += strlen(p)+1;
  }

  if (set) {
    u_int32 t = (u_int32)mktime(tm);
    if (tim < 0 && set < 3) {
      printf("E: RTC not set. You need to set both date and time!\n");
      goto finally;
    }
    if (SetRtc(t) != t) {
      printf("E: RTC not found. Check RTC battery!\n");
      goto finally;
    }
  }

  res = S_OK;
 finally:
  return res;
}





#if 0
char *wday[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

void PrintTime(time_t tt) {
  struct tm *tm = localtime(&tt);
  printf("  %04d-%02d-%02d at %02d:%02d:%02d, d=%s, yd=%d, dst=%d\n",
	 tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
	 tm->tm_hour, tm->tm_min, tm->tm_sec,
	 wday[tm->tm_wday], tm->tm_yday, tm->tm_isdst);
}
#endif



DLLENTRY(main)
ioresult main(char *parameters) {
  int nParam = RunLib("ParamSpl", parameters);
  char *p = parameters;
  int autoPrint = 1;
  int i, res = S_ERROR;
  int set = 0;
  static char s[256];
  time_t tim = time(NULL);
  struct tm *tm = localtime(&tim);

  if (!strcmp(p, "-s")) {
    res = SetDate(p+3, nParam-1);
    goto finally;
  }

  /* If not help or set mode... */
  if (strcmp(p, "-h")) {
    if (tim == TIME_NOT_FOUND) {
      printf("ERROR! Could not find RTC!\n"
	     "Please check that you have put an LR44 battery to the\n"
	     "BATT1 connector on the VS1005g developer board, or\n"
	     "otherwise powered RTC!\n");
      goto finally;
    } else if (tim == TIME_NOT_SET) {
      printf("ERROR! RTC not set. Please run Date -s YY-MM-DD HH:MM:SS\n"
	     "or C function SetRtc() (e.g. SetRtc(mktime(tm))) to set RTC!\n");
      goto finally;
    }
  }

 //printf("nPAram: %x\n", nParam);
 //printf("PP:%s\n", p);

  for (i=0; i<nParam; i++) {
   // printf("P:%s\n", p);
    if (!strcmp(p, "-h")) {
      printf("Usage: Date [formatString] [-t|-q|-h]\n"
	     "       Date -s [[YY]YY-MM-DD|HH:MM[:SS]]\n"
	     "-t\tPrint uptime\n"
	     "-q\tQuiet mode\n"
	     "-h\tShow this help\n"
	     "-s\tSet date and/or time\n\n"
	     "FormatString:\n"
	     "%%a wdayName   %%A weekdayName   %%b monName    %%B monthName\n"
	     "%%c date&time  %%d dayOfMonth    %%H hour-24    %%h hour-12\n"
	     "%%j dayOfYear  %%m month         %%M minute     %%p AM/PM\n"
	     "%%S second     %%w weekDay,S=0   %%x date       %%X time\n"
	     "%%y yr         %%Y year          %%Z UTC        %%%% %%\n\n"
	     "[YY]YY-MM-DD\tSet date, e.g. 2019-05-13 or 19-05-13\n"
	     "HH:MM[:SS]\tSet time, e.g. 12:34:56 or 12:34\n"
	     );
      res = S_OK;
      goto finally;
    } else if (!strcmp(p, "-q")) {
      res = S_OK;
      goto finally;
    } else if (!strcmp(p, "-t")) {
      if (!autoPrint) {
	putchar(' ');
      }
      printf("%5.3fs", ReadTimeCount()*0.001);
      autoPrint = 0;
    } else {
      if (!autoPrint) {
	putchar(' ');
      }
      strftime(s, 256, p, localtime(&tim));
      fputs(s, stdout);
      autoPrint = 0;
    }
    p += strlen(p)+1;
  }

  if (autoPrint) {
    strftime(s, 256, "%c", localtime(&tim));
    puts(s);
  } else {
    putchar('\n');
  }

  res = S_OK;
 finally:
  return res;
}

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

#define leap(y) (((y) & 3) == 0)

const char * __mem_y wdayNames[7] = {
  "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
};

const char * __mem_y monNames[12] = {
  "January", "February", "March", "April", "May", "June",
  "July", "August", "September", "October", "November", "December"
};

const char * __mem_y digFmt[5] = {
  "%d", "%1d", "%02d", "%03d", "%04d"
};

/* Helper function for strftime() */
char *StrfCondPrintS(register char *d, register size_t *smax,
		     register s_int16 n, register const char *s) {
  while (*smax && n && *s) {
    *d++ = *s++;
    (*smax)--;
    n--;
  }
  return d;
}

/* Helper function for strftime() */
char *StrfCondPrint(register char *d, register size_t *smax,
		   register s_int16 digits, register s_int16 val) {
  char s[5];
  sprintf(s, digFmt[digits], val);
  return StrfCondPrintS(d, smax, digits, s);
}


size_t strftime(register char *d, register size_t smax,
		register const char *fmt, register const struct tm *tp) {
  int esc=0;
  int c;
  size_t origSMax = --smax;
  if (smax < 0) {
    return 0;
  }
  while (smax > 0 && (c = *fmt++)) {
    size_t smaxCopy = smax;
    if (esc) {
      switch(c) {
      case 'a':
	d = StrfCondPrintS(d, &smaxCopy, 3, wdayNames[tp->tm_wday]);
	break;
      case 'A':
	d = StrfCondPrintS(d, &smaxCopy, 255, wdayNames[tp->tm_wday]);
	break;
      case 'b':
	d = StrfCondPrintS(d, &smaxCopy, 3, monNames[tp->tm_mon]);
	break;
      case 'B':
	d = StrfCondPrintS(d, &smaxCopy, 255, monNames[tp->tm_mon]);
	break;
      case 'c':
	{
	  char t[20];
	  sprintf(t, "%04d-%02d-%02d %02d:%02d:%02d",
		  tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday,
		  tp->tm_hour, tp->tm_min, tp->tm_sec);
	  d = StrfCondPrintS(d, &smaxCopy, 19, t);
	}
	break;
      case 'd':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_mday);
	break;
      case 'H':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_hour);
	break;
      case 'h':
	{
	  int hour = tp->tm_hour % 12;
	  d = StrfCondPrint(d, &smaxCopy, 2, hour ? hour : 12);
	}
	break;
      case 'j':
	d = StrfCondPrint(d, &smaxCopy, 3, tp->tm_yday+1);
	break;
      case 'm':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_mon+1);
	break;
      case 'M':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_min);
	break;
      case 'p':
	d = StrfCondPrintS(d, &smaxCopy, 2, (tp->tm_hour < 12) ? "AM" : "PM");
	break;
      case 'S':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_sec);
	break;
      case 'U':
      case 'W':
	d = StrfCondPrintS(d, &smaxCopy, 2, "-1");
	break;
      case 'w':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_wday);
	break;
      case 'x':
	{
	  char t[11];
	  sprintf(t, "%04d-%02d-%02d",
		  tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday);
	  d = StrfCondPrintS(d, &smaxCopy, 10, t);
	}
	break;
      case 'X':
	{
	  char t[9];
	  sprintf(t, "%02d:%02d:%02d",
		  tp->tm_hour, tp->tm_min, tp->tm_sec);
	  d = StrfCondPrintS(d, &smaxCopy, 8, t);
	}
	break;
      case 'y':
	d = StrfCondPrint(d, &smaxCopy, 2, tp->tm_year % 100);
	break;
      case 'Y':
	d = StrfCondPrint(d, &smaxCopy, 4, tp->tm_year+1900);
	break;
      case 'Z':
	d = StrfCondPrintS(d, &smaxCopy, 3, "UTC");
	break;
      default:
	*d++ = c;
	smax--;
	break;
      }
      esc = 0;
    } else if (c == '%') {
      esc = 1;
    } else {
      *d++ = c;
      smax--;
    }
    smax = MIN(smax, smaxCopy);
  }
  *d++ = '\0';
  return origSMax - smax;
}

struct tm *localtime(register const time_t *tp) {
  static struct tm tm;
  time_t t = *tp, nextT;

  /* Quick and dirty implementation. Shouldn't reinvent the wheel, but
     here we go. */
  memset(&tm, 0, sizeof(tm));
  if (t < 0 && t > -128) {
    /* Invalid clock; return 1999-31-12 12:00:00 */
    tm.tm_year = 99;
    tm.tm_mon = 12-1;
    tm.tm_mday = 31;
    tm.tm_yday = 364;
    tm.tm_hour = 12;
    goto finally;
  }
  tm.tm_wday = ((u_int16)(t/86400) + 6) % 7;

  tm.tm_year = 100;
  while ((nextT = t - (365+leap(tm.tm_year+1900))*86400) >= 0) {
    //    printf("    year %d, nextT %ld, t %ld\n", tm.tm_year+1900, nextT, t);
    tm.tm_year++;
    t = nextT;
  }
  tm.tm_yday = (int)(t/86400);

  while ((nextT = t - monLen[leap(tm.tm_year+1900)][tm.tm_mon]*86400) >= 0) {
    //    printf(" mon %d, nextT %ld, t %ld\n", i, nextT, t);
    tm.tm_mon++;
    t = nextT;
  }

  tm.tm_mday = (int)(t/86400);
  t -= tm.tm_mday*86400;

  tm.tm_hour = (int)(t/3600);
  t -= tm.tm_hour*3600;

  tm.tm_min = (int)t/60;
  tm.tm_sec = (int)t - tm.tm_min*60;

  tm.tm_mday++;

 finally:
  return &tm;
}

time_t time(register time_t *tp) {
  time_t t = GetRtc();
  if (tp) {
    *tp = t;
  }
  return t;
}

time_t mkgmtime (register struct tm *tm);

/* Return the equivalent in seconds past 12:00:00 a.m. Jan 1, 1970 GMT
   of the local time and date in the exploded time structure `tm',
   and set `tm->tm_yday', `tm->tm_wday', and `tm->tm_isdst'.
   Return -1 if any of the other fields in `tm' has an invalid value. */

time_t mktime (register struct tm *tm) {
  struct tm save_tm;            /* Copy of contents of `*tm'. */
  struct tm *ltm;               /* Local time. */
  time_t then;                  /* The time to return. */

  then = mkgmtime (tm);
  if (then == -1)
    return -1;

#if 0
  /* In case `tm' points to the static area used by localtime,
     save its contents and restore them later. */
  save_tm = *tm;
  /* Correct for the timezone and any daylight savings time.
     If a change to or from daylight savings time occurs between when
     it is the time in `tm' locally and when it is that time in Greenwich,
     the change to or from dst is ignored, but that is a rare case. */
  then += then - mkgmtime (localtime (&then));

  ltm = localtime (&then);
  save_tm.tm_yday = ltm->tm_yday;
  save_tm.tm_wday = ltm->tm_wday;
  save_tm.tm_isdst = ltm->tm_isdst;
  *tm = save_tm;
#endif

  return then;
}

/* Nonzero if `y' is a leap year, else zero. */
/* Valid between 1901 to 2099, which is enough for us */
#define leap(y) (((y) & 3) == 0)

/* Number of leap years from 1970 to `y' (not including `y' itself). */
/* Valid between 1901 to 2099, which is enough for us */
#define nleap(y) (((y) - 1997) / 4)


/* Number of days in each month of the year. */
static char __mem_y monlens[] =
{
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

/* Return the equivalent in seconds past 12:00:00 a.m. Jan 1, 2000 GMT
   of the Greenwich Mean time and date in the exploded time structure `tm',
   and set `tm->tm_yday', `tm->tm_wday', and `tm->tm_isdst'.
   Return -1 if any of the other fields in `tm' has an invalid value. */

time_t mkgmtime (register struct tm *tm) {
  int years, months;
  volatile int days;
  time_t res;

  years = tm->tm_year + 1900;   /* year - 1900 -> year */
  months = tm->tm_mon;          /* 0..11 */
  days = tm->tm_mday - 1;       /* 1..31 -> 0..30 */

  if (years < 2000
      || months < 0 || months > 11
      || days < 0
      || days > monlens[months] + (months == 1 && leap (years)) - 1
      || tm->tm_hour < 0 || tm->tm_hour > 23
      || tm->tm_min < 0 || tm->tm_min > 59
      || tm->tm_sec < 0 || tm->tm_sec > 61) {
    return -1;
  }
  
  res = 3600 * (s_int32)tm->tm_hour + (60 * tm->tm_min + tm->tm_sec);

  /* Set `days' to the number of days into the year. */
  if (months > 1 && leap (years))
    ++days;
  while (months-- > 0)
    days += monlens[months];
  tm->tm_yday = days;

  /* Now set `days' to the number of days since Jan 1, 2000. */
  days += 365 * (years - 2000) + nleap (years);
#if 0
  printf("#DAYS%d#HR%d#MIN%d#SEC%d#", days, tm->tm_hour, tm->tm_min, tm->tm_sec);
#endif
  tm->tm_wday = (days + 6) % 7; /* Jan 1, 2000 was Saturday. */
  tm->tm_isdst = 0;

  res +=  (s_int32)days * 86400;

  return res;
}

int __mem_y monLen[2][12] = {
  {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
  {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
};





