#include <string.h>
#include <stdlib.h>
#include <voaudio.h>

// #include "arcstruct.h"
#include "fatdircache.h"
#include "lcdtimer.h"

// #define WITH_ARC_LCD

#ifndef ARCSTRUCT_H
__y DirCache dirc; 
#endif

__x DirNameAndCluster dirs[DIR_LIST_ENTRIES];
DirNameAndCluster *dirListTop = dirs; //Directory Sort Pointer Global 

__y char *fnsp = fileNameStash;
__y char fileNameStash[9000];

/// Print current contents of the directory cache dirc.
void PrintDirectoryCache() {
	__y DirCacheEntry *dc = dirc.e;
	int i;
	printf("\n == %d entries in cache ==\n",dirc.n);
	for (i=0; i<dirc.n; i++) {
		printf("%c ",(dirc.cur == &dirc.e[i]) ? '>' : ' ');
		printf("%3d[%p]: 1st song %5d at %6lx '%s'\n",i,&dirc.e[i],dirc.e[i].firstSongNumber,dirc.e[i].dirStartCluster,IdentifyDirectory(dirc.e[i].dirStartCluster));
		dc++;
	}
	printf("====\n");
}



void PrintSpaces(int n) {
	while(n--) printf(" ");
}

void PrintYString(__y char *s) {
	while (*s) printf("%c",*s++);
}



/// Print current contents of the current directory sort array dirs.
void PrintDirectorySortArray() {
	DirNameAndCluster *dirListTop = dirs;
	int i;
	for (i=0;;i++) {
		printf("[%p]%4d %7lx [%p]:",dirListTop,i,dirListTop->dirStartCluster,dirListTop->dirName);
		PrintYString(dirListTop->dirName);			
		printf("\n");
		dirListTop++;
		if (!dirListTop->dirStartCluster) break;
	}
}


/// Try to insert the current sorted directory list 'dirs' to the directory cache 'dirc' to the location pointed by dirc.cur. 
/// Updates numberOfDirectories.
/// \return Returns 1 if subdirectory fits into the dirc list
int InsertSortListToCache(DirNameAndCluster *bottom, int numberOfDirectories) {
	int i;
	//printf("inserting %d directories\n",numberOfDirectories);
	//PrintDirectorySortArray();
	if (numberOfDirectories < (dirc.total - dirc.n)) { //There's size for this directory in the global directory cache
		__y DirCacheEntry *c = dirc.cur;
		memmoveYY(&c[numberOfDirectories], c, sizeof(c[0])*dirc.n);
		for (i=0; i < numberOfDirectories; i++) {
			c->dirStartCluster = bottom[i].dirStartCluster;
			//printf("(%lx)",bottom[i].dirStartCluster);
			c->firstSongNumber = 0; //Unknown, to be populated later
			c++;
		}
		dirc.n += numberOfDirectories;
		//PrintDirectoryCache();
		return 1;
	}
	printf("No space in cache!\n");
	return 0;
}


int IsPlayableFile(DirectoryEntry *de) {
	if (de->attr & (__ATTR_DIRECTORY | __ATTR_VOLUMEID | __ATTR_HIDDEN | __ATTR_SYSTEM)) return 0;
	if (de->filename[4] != (('M' << 8) | 'P')) return 0;
	if ((de->attr >> 8) != '3') return 0;
	return 1;
}


// Adds directory to sort list dirs if the file name can fit into the fileNameStash. Otherwise does not do anything.
void AddDirectoryToSortList(DirectoryEntry *de) {
	u_int32 firstCluster = ((u_int32)BYTESWAP(de->firstClusHi) << 16) | BYTESWAP(de->firstClusLo);
	if (IsPlayableFile(de)) dirc.nextSongNumber++;
	if (dirListTop >= &dirs[sizeof(dirs)]) return;
	if ((longFileName[0] != '.') && ((de->attr & __ATTR_DIRECTORY) != 0) && ((de->attr & __ATTR_SYSTEM) == 0) && ((de->attr & __ATTR_VOLUMEID) == 0)) {
		int n = strlen(longFileName) + 1;
		if (fnsp+n < &(fileNameStash[sizeof(fileNameStash)])) {
			memcpyXY(fnsp,longFileName,n); //Stash name of directory (trailing zero included) to *fnsp
			dirListTop->dirName = fnsp;
			dirListTop->dirStartCluster = firstCluster;
			dirListTop++;
			fnsp += n;			
		}
	}
}


// Adds directory to sort list dirs if the file name can fit into the fileNameStash. Otherwise does not do anything.
void AddSongToSortList(DirectoryEntry *de) {
	u_int32 firstCluster = ((u_int32)BYTESWAP(de->firstClusHi) << 16) | BYTESWAP(de->firstClusLo);
	if (dirListTop >= &dirs[sizeof(dirs)]) return;
	if (IsPlayableFile(de)) {
		int n = strlen(longFileName) + 1;
		if (fnsp+n < &(fileNameStash[sizeof(fileNameStash)])) {
			memcpyXY(fnsp,longFileName,n); //Stash name of directory (trailing zero included) to *fnsp
			dirListTop->dirName = fnsp;
			dirListTop->dirStartCluster = firstCluster;
			dirListTop++;
			fnsp += n;			
		}
	}
}



	
/// Compare function for directory sorting
int CompareByDirectoryCluster(DirNameAndCluster *a, DirNameAndCluster *b) {
	if (a->dirStartCluster == b->dirStartCluster) return 0;
	if (a->dirStartCluster < b->dirStartCluster) return -1;
	return 1;	
}

/// Compare function for directory sorting
int CompareByDirectoryName(DirNameAndCluster *a, DirNameAndCluster *b) {
	__y char *ca = a->dirName;
	__y char *cb = b->dirName;
	while(*cb) {
		if (*ca < *cb) return -1;
		if (*ca > *cb) return 1;
		ca++;
		cb++;
	}
	return *ca; //zero if equal length, otherwise positive for latin characters.
}


/// Build a sort list (dirs) from a directory file.
int ReadAndSortDirectoryFile(FILE *f, DirNameAndCluster *dirListBuffer) { //Reads and sorts a directory. FILE *f must point to beginning of a FAT directory (the "." entry).
	int n;
	dirListTop = dirListBuffer;
	fnsp = fileNameStash;
	ScanDirectoryFile(f, AddDirectoryToSortList);
	n = (dirListTop-dirListBuffer);
	qsort(dirListBuffer,n,sizeof(dirs[0]),(void*)CompareByDirectoryName);
	return n;
}


/// Open a directory file from directory start cluster and build a sort list (dirs) from it.
u_int16 ReadAndSortDirectoryCluster(u_int32 dirStartCluster, DirNameAndCluster *dirListBuffer) {
	u_int16 n;
	FILE *f = dopen_fat(dirStartCluster);
	FatFileInfo *fi=(FatFileInfo*)f->fileInfo;
	if (!f) return;
	n = ReadAndSortDirectoryFile(f, dirListBuffer);
	fclose(f);
	return n;
}

/// Initialize the directory cache
void DirCacheInit() {
	dirc.n = 1; //Root is implied
	dirc.nextSongNumber = 1;
	dirc.total = sizeof(dirc.e)/sizeof(dirc.e[0]);
	memsetY(dirc.e, 0, dirc.total * sizeof(DirCacheEntry));
	dirc.cur = dirc.e;
}


void leaf(u_int32 dirStartCluster, DirNameAndCluster *dirListBuffer) {
	int i;
	int n = ReadAndSortDirectoryCluster(dirStartCluster, dirListBuffer);
	for (i=0; i<n; i++) {
		leaf(dirListBuffer[i].dirStartCluster, &dirListBuffer[n]);
	}
}
	

extern u_int32 t1,t2,t3;

#ifdef WITH_ARC_LCD
#include "lcdui.h"
#endif

void RunLeafs() {
	int i;
	for (i=0; i<dirc.n; i++) { 
		t3 = ReadTimeCount();
		if (t3-t2 > 20) {
			t2 = ReadTimeCount();
			LcdPutUnsigned4(dirc.nextSongNumber-1);
			printf("\r%5d",dirc.nextSongNumber-1);
		}
			
		if (dirc.e[i+1].firstSongNumber == 0) {
			dirc.e[i].firstSongNumber = dirc.nextSongNumber;
			leaf(dirc.e[i].dirStartCluster,dirs);
		}		
	}
}






void BuildDirectoryCache() {
	DirCacheInit();	
	do {
		dirc.cur->firstSongNumber = dirc.nextSongNumber;
		ReadAndSortDirectoryCluster(dirc.cur->dirStartCluster, dirs);
		dirc.cur++;
		if (!InsertSortListToCache(dirs,dirListTop-dirs)) { //Returns 1 if subdirectory fits into the dirc list
			RunLeafs();
			break;
		}
		LcdPutUnsigned4(dirc.nextSongNumber-1);
	} while(dirc.cur->dirStartCluster);
	//dirc.cur->firstSongNumber = dirc.nextSongNumber;
	dirc.totalSongs = dirc.nextSongNumber-1;
	if ((dirc.nowPlaying == 0) || (dirc.nowPlaying > dirc.totalSongs)) dirc.nowPlaying = 1;
}


