#include <vo_stdio.h>
#include <vo_fat.h>
#include <string.h>
#include <swap.h>
#include <stdlib.h>
#include "fatdirectory.h"
#include "lcdtimer.h"
#include <ctype.h>


char *currentDisk = "M:";
	


// Include functions to open a song by number
#define WITH_OPEN_NUMBERED_SONG
u_int16 leafFindSongNumber; 
u_int16 leafFirstSongInDirectory;
u_int32 songStartCluster;
u_int32 songDirectoryCluster;
u_int32 parentDirectoryCluster = 0;

// Include LCD functionality to draw file names on the ARC LCD screen
// #define WITH_ARC_LCD
// #include "arcstruct.h"
#include "lcdui.h"




// Open directory as a file. Takes directory start cluster as parameter (0 means root directory). currentDisk must point to diskspec such as "S:"
FILE *dopen_fat(u_int32 dirStartCluster) {
	FILE *f = fopen(currentDisk,"s");
	FatDeviceInfo *di = f->dev->deviceInfo;
	FatFileInfo *fi = f->fileInfo;
	

	if (f) {
		FatFileInfo *fi=(FatFileInfo*)f->fileInfo;
		memset(fi, 0, 11);
		f->fileSize = 1000000000;
			
		if (dirStartCluster == 0) {
			dirStartCluster = *(u_int32*)(&(f->dev->deviceInfo[10]));
			//printf("Root->%ld ",dirStartCluster);
			if (di->fatBits != 32) {
				fi->currentFragment.startSector = di->rootStart;
				f->fileSize = fi->currentFragment.sizeBytes = (u_int32)di->rootEntCnt*32;
			}
		} else {
			//printf("dopen->%ld ",dirStartCluster);
		}
	
		fi->startCluster = dirStartCluster;

	}
	return f;
}




/// Scan one FAT directory entry. Builds long file name as it goes and calls EntryHandler function pointer for each real entry
int ScanDirectoryEntry(DirectoryEntry *de, void (*EntryHandler)(DirectoryEntry *de)) {
	char tempName[13];
	if (!(de->filename[0] & 0xff00)) return 0; //End of files in directory
	if ((de->filename[0]>>8) == 0xe5) { //this is a deleted entry
	} else if ((de->attr & __ATTR_LONGNAME) != __ATTR_LONGNAME) {
		FatNameFromDirEntry((u_int16*)de, tempName); //short name to tempName
		if (longFileName[0]==0) strcpy(longFileName,tempName); //If no longFileName, set it from short file name		
		EntryHandler(de);		
		longFileName[0] = 0;
	} else { //longname entry
		u_int16 fn = GetByte((void*)de,0);
		register u_int16 idx = ((fn & 0x3f) - 1) * 13;
		static const short places[] = {	1, 3, 5, 7, 9, 14, 16, 18, 20, 22, 24, 28, 30, -1 };
		register const short *p = &places[0];
		while (*p >= 0 && idx < FAT_LFN_SIZE-1) { longFileName[idx++] =	(GetByte((void*)de, *p+1) << 8) + GetByte((void*)de, *p); p++; }
		if (fn & 0x40) { //last long name entry
			if (idx >= FAT_LFN_SIZE) idx = FAT_LFN_SIZE-1;
			longFileName[idx] = '\0';
		}
	}
	return 1;
}



// Scan a directory file. Calls EntryHandler for all real entries in the directory file.
void ScanDirectoryFile (FILE *f, void (*EntryHandler)(DirectoryEntry *de)) {
	DirectoryEntry de;
	u_int16 i;
	longFileName[0] = 0;
	f->fileSize = 1000000; // VSOS fails to populate fileSize with correct length of directory when opening a directory as a file.
	do {
		if (!fread (&de,sizeof(de),1,f)) break;
		//PrintDirectoryEntry(&de);
	} while (ScanDirectoryEntry(&de, EntryHandler));
}





// Prints contents of directory entry in human-readable form.
void PrintDirectoryEntry(DirectoryEntry *de) {
	u_int16 *buf = (u_int16*)de;
	s_int16 i, j;
	char tempName[13];
	for (j=0; j<2; j++) {
			printf("   ");
		for (i=0; i<16; i++) {
			printf("%04x ",buf[j*16+i]);
		}
		for (i=0; i<16; i++) {
			char ch=buf[j*16+i] >> 8;
			if (ch < 32) ch = '.';
			if (ch > 127) ch = '.';
			putchar(ch);
			ch=buf[j*16+i] & 0xff;
			if (ch < 32) ch = '.';
			if (ch > 127) ch = '.';
			putchar(ch);
		}
		printf("\n");
	}
	if (!(de->filename[0] & 0xff00)){
		printf("End of directory\n");
	} else if ((de->filename[0]>>8) == 0xe5) { //this is a deleted entry
		printf("Deleted entry\n");
	} else if ((de->attr & __ATTR_LONGNAME) != __ATTR_LONGNAME) {
		u_int16 fdate;
		u_int32 fileSize = ((u_int32)BYTESWAP(de->fileSizeHi) << 16) | BYTESWAP(de->fileSizeLo) ;
		FatNameFromDirEntry((u_int16*)de, tempName); //short name to tempName
		printf("Main entry: %s, %ld bytes. ",tempName,fileSize);		
		fdate = Swap16(de->creatDate);
		printf("Created %d-%d-%d, ",(fdate>>9)+1980, (fdate>>5)&0xf, (fdate)&0x1f);
		fdate = Swap16(de->writeDate);
		printf("Modified %d-%d-%d, ",(fdate>>9)+1980, (fdate>>5)&0xf, (fdate)&0x1f);
		fdate = Swap16(de->lastAccessDate);
		printf("Accessed %d-%d-%d\n",(fdate>>9)+1980, (fdate>>5)&0xf, (fdate)&0x1f);				
	} else { //longname entry
		printf("Long name entry\n");
	}
}


u_int32 matchCluster = 0;
char identifiedDirectory[124];
u_int32 identifiedFileSize;

void IdentifyDirectoryMatchHandler(DirectoryEntry *de) {
	u_int32 firstCluster = ((u_int32)BYTESWAP(de->firstClusHi) << 16) | BYTESWAP(de->firstClusLo);
	if (firstCluster == matchCluster) {
		strcpy(identifiedDirectory,longFileName);
		matchCluster = 0;
		identifiedFileSize =((u_int32)BYTESWAP(de->fileSizeHi) << 16) | BYTESWAP(de->fileSizeLo) ;
	}	
}
	

char *IdentifyDirectory(u_int32 dirStartCluster) {
	FILE *f = dopen_fat(dirStartCluster);
	DirectoryEntry de;
	f->pos = 32;
	fread (&de,sizeof(de),1,f);
	fclose(f);
	parentDirectoryCluster = (((u_int32)BYTESWAP(de.firstClusHi) << 16) | BYTESWAP(de.firstClusLo));
	if (de.filename[0] != (('.' << 8) | '.')) {
		parentDirectoryCluster = 0;
		return ("Root");
	}
	f = dopen_fat(((u_int32)BYTESWAP(de.firstClusHi) << 16) | BYTESWAP(de.firstClusLo));
	matchCluster = dirStartCluster;
	while(matchCluster) {
		fread (&de,sizeof(de),1,f);
		ScanDirectoryEntry(&de,IdentifyDirectoryMatchHandler);
	}
	fclose(f);
	return identifiedDirectory;
}



// Stuff before this is pretty generic. But let's add a couple of functions that help with MP3 players specifically, to open songs based on a number.
// For this, we need to use also the fatdircache routines and fatdircache memory.
#ifdef WITH_OPEN_NUMBERED_SONG
#include "fatdircache.h"

// Populate file descriptor with correct file size (after opening by cluster)
void GetOpenedSongFileSize(VO_FILE *song) {
	VO_FILE *f = dopen_fat(songDirectoryCluster);
	matchCluster = songStartCluster;	
	ScanDirectoryFile(f, IdentifyDirectoryMatchHandler);
	fclose(f);	
	song->fileSize = identifiedFileSize;
}


/// Find a directory which contains the song number leafFindSongNumber, starting from a directory cluster dirStartCluster.
u_int32 leafFindSongDirectory(u_int32 dirStartCluster, DirNameAndCluster *dirListBuffer) {
	u_int16 i,n;
	u_int32 d;
	leafFirstSongInDirectory = dirc.nextSongNumber;
	n = ReadAndSortDirectoryCluster(dirStartCluster, dirListBuffer);	
	if (dirc.nextSongNumber > leafFindSongNumber) return dirStartCluster;	
	for (i=0; i<n; i++) if (d = leafFindSongDirectory(dirListBuffer[i].dirStartCluster, &dirListBuffer[n])) return d;
	return 0;
}


int FirstSongOfNextAlbum(u_int16 songNumber) {
	int i = 0;
	int n;
	// Find the last directory cache entry whose firstSongNumber is below or equal to the song number we are looking for
	while (dirc.e[i].firstSongNumber && (dirc.e[i].firstSongNumber <= songNumber) && (i<dirc.n-1)) i++;
	n = dirc.e[i].firstSongNumber;
	//printf("Find: %d, dirc[%d]=%d\n",songNumber,i,n);
	return n;
}

extern char songFileName[FAT_LFN_SIZE];

/// FAST open a song by number by doing a directory cache assisted sorted depth-first search.
FILE *OpenSongByNumber(u_int16 songNumber) {
	int i = 0;
	int n;
	FILE *f;
	printf(">%d.",songNumber);

	// Find the last directory cache entry whose firstSongNumber is below or equal to the song number we are looking for
	while (dirc.e[i].firstSongNumber && (dirc.e[i+1].firstSongNumber <= songNumber) && (i<dirc.n-1)) i++;
	// Populate dirc algorithm globals
	leafFindSongNumber = songNumber;
	dirc.nextSongNumber = dirc.e[i].firstSongNumber;
	// Depth first search through subdirectories starting from the cached directory entry until we find the correct directory
	songDirectoryCluster = leafFindSongDirectory(dirc.e[i].dirStartCluster, dirs);
	// Calculate the number of the wanted file in the depth first searched directory
	i = songNumber - leafFirstSongInDirectory;
	// Read the directory to dirs array of file names and sort the array
	f = dopen_fat(songDirectoryCluster);	
	dirListTop = dirs;
	fnsp = fileNameStash;
	ScanDirectoryFile(f, AddSongToSortList);
	fclose(f);	
	n = (dirListTop-dirs);
	qsort(dirs,n,sizeof(dirs[0]),(void*)CompareByDirectoryName);	
	// dirs is now sorted, thus dirs[i] is the song we need.

	/*
	printf("\n");
	if (parentDirectoryCluster) {
		printf("%s/",IdentifyDirectory(parentDirectoryCluster));
	}
	printf("%s/",IdentifyDirectory(songDirectoryCluster));
	*/

	memset(songFileName,' ',FAT_LFN_SIZE); //150 words
	{
		static int wasLower;
		static int wasDigit;
		char *p = &songFileName[3];
		wasLower = 0;
		wasDigit = 0;
		for (n=0; n<FAT_LFN_SIZE-20; n++) {
			char ch = dirs[i].dirName[n]; //file name
			if (!ch) break; //end of string
			printf("%c",ch);
			if (ch == '.') break; //end of filename
			if (ch == '\'') continue;
			if ((wasLower) && (isupper(ch))) *p++ = ' ';
			if ((wasLower) && (isdigit(ch))) *p++ = ' ';
			if ((wasDigit) && (isalpha(ch))) *p++ = ' ';
			wasLower = islower(ch);
			wasDigit = isdigit(ch);
			ch = toupper(ch);
			if (ch == '_') ch = ' ';
			if (ch == '-') ch = ' ';			
			if ((isalnum(ch)) || (ch==' ')) {
				if (ch == 'W') {
					*p++ = 'U';
					*p++ = 'W';
				} else if (ch == 'M') {
					*p++ = 'N';
					*p++ = 'M';
				} else if (ch == 'X') {
					*p++ = 'K';
					*p++ = 'S';
				} else {
					*p++ = ch;
				}
			}
		}
		p[4] = 0;
	}
	printf("[%s]",songFileName);
	
	// Open the song file.
	songStartCluster = dirs[i].dirStartCluster;
	f = dopen_fat(songStartCluster);
	GetOpenedSongFileSize(f);
	//printf(" %ld bytes ",f->fileSize);
	return f;
}



#endif

