
#include <vo_stdio.h>
#include <vs1010c.h>
#include "lcdtimer.h"
#include <xvm.h>
#include <ctype.h>



/// \file main.c VS1010D VSOS Executable main file
/// This is a starting point for developing VS1010D DLX libraries and executables

#include <vo_stdio.h>
#include <volink.h>     // Linker directives like DLLENTRY
#include <apploader.h>  // RunLibraryFunction etc
#include <vs1010dRom.h>
#include <vo_gpio.h>
#include <vs1010c.h>
#include <playerinfo.h>
#include <string.h>
#include <protocol.h>
#include <spitv.h>
#include <lcd.h>
#include <stdlib.h>


#define DO_BRUTAL_BUTTON_INPUT
// "Brutal" button input writes raw values to GPIO DDR registers.
// If GPIO pin usage changes, the below values must change accordingly:

#define GPIO0_DDR_BEFORE_BUTTON_READ 0x0071 
#define GPIO1_DDR_BEFORE_BUTTON_READ 0x000d
#define GPIO2_DDR_BEFORE_BUTTON_READ 0x3000

#define GPIO0_DDR_AFTER_BUTTON_READ 0x0079
#define GPIO1_DDR_AFTER_BUTTON_READ 0x00fd
#define GPIO2_DDR_AFTER_BUTTON_READ 0x72f1

// The idea is that the SEG pins will switch to output after each LCD frame, 
// then there's 1 idle step to allow signals to settle, then they are read back to gpioIn[3],
// from where their button pressed/notpressed states can be read.
// The GPIO_DDR values that switch the pins to inputs, and then back to outputs has been precalculated above to make it fast. But it's very error prone if GPIO usage - ANY gpio usage - changes in the product. 



u_int16 lcdd[4] = {0};
u_int16 lcdCommons[4] = {0x20,0x21,0x22,0x23};
u_int16 lcdSegments[16] = {0x2e, 0x24, 0x25, 0x26, 0x27, 0x14, 0x15, 0x16, 0x17, 0x03, 0x29, 0};
volatile u_int16 lcdPhase = 0;
const u_int16 phaseCommon[9] = {(1<<0),(1<<1),(1<<2),(1<<3),(1<<0),(1<<1),(1<<2),(1<<3),0}; //Active common for each phase
u_int16 phaseGpioHigh[9*3] = {0};
u_int16 phaseGpioLow[3];
u_int16 gpioIn[3];

int GetButtonPress() {
	int i;
	u_int16 s;
	for (i=0; s = lcdSegments[i]; i++) {
		if (gpioIn[s>>4] & (1 << (s&0xf))) return i+1;		
	}
}


void LcdGenerate() {
	int ph,i;
	u_int16 phaseGpioHigh2[9*3];
	memset(phaseGpioHigh2, 0, sizeof(phaseGpioHigh2));
	for (ph=0; ph<8; ph++) {
		int com = ph & 3;
		int inv = ph & 4;
		for (i=0; i<16; i++) {
			int k = lcdd[com] & (1<<i);
			int s = lcdSegments[i];
			if (inv) k = !k;
			if (s && k) phaseGpioHigh2[ph*3 + (s>>4)] |= (1 << (s&0xf));
			if (s) phaseGpioLow[s>>4] |= (1 << (s&0xf));
		}			
	}
	memcpy(phaseGpioHigh, phaseGpioHigh2, sizeof(phaseGpioHigh));
}






void LcdStep() {
	u_int16 v;
	u_int16 k;

	lcdPhase++;
	if (lcdPhase >= 9) lcdPhase = 0;

	// set commons
	v = PERIP(GPIO2_DDR) & ~(0xf); // Zero(input) for pins 0..3
	k = phaseCommon[lcdPhase];
	PERIP(GPIO2_DDR) = v | k; // one common is output
	if (lcdPhase & 4) {
		PERIP(GPIO2_SET_MASK) = k;
	} else {
		PERIP(GPIO2_CLEAR_MASK) = k;
	}

	// set segments from precalculated table
	{
		u_int16 *p = &phaseGpioHigh[lcdPhase * 3];
		PERIP(GPIO0_CLEAR_MASK) = phaseGpioLow[0];
		PERIP(GPIO1_CLEAR_MASK) = phaseGpioLow[1];
		PERIP(GPIO2_CLEAR_MASK) = phaseGpioLow[2];		
		PERIP(GPIO0_SET_MASK) = *p++;
		PERIP(GPIO1_SET_MASK) = *p++;
		PERIP(GPIO2_SET_MASK) = *p++;
	}	
	

#ifdef DO_BRUTAL_BUTTON_INPUT
	if (lcdPhase == 8) {
		PERIP(GPIO0_DDR) = GPIO0_DDR_BEFORE_BUTTON_READ;	
		PERIP(GPIO1_DDR) = GPIO1_DDR_BEFORE_BUTTON_READ;	
		PERIP(GPIO2_DDR) = GPIO2_DDR_BEFORE_BUTTON_READ;	
	}
	if (lcdPhase == 0) {
		gpioIn[0] = PERIP(GPIO0_IDATA);
		gpioIn[1] = PERIP(GPIO1_IDATA);
		gpioIn[2] = PERIP(GPIO2_IDATA);
		PERIP(GPIO0_DDR) = GPIO0_DDR_AFTER_BUTTON_READ;	
		PERIP(GPIO1_DDR) = GPIO1_DDR_AFTER_BUTTON_READ;	
		PERIP(GPIO2_DDR) = GPIO2_DDR_AFTER_BUTTON_READ;	
	}
#endif

}
	
void LcdSetAllSignals(int hilo) {
	int i;
	// set commons;
	for (i=0; i<4; i++) {
		if (lcdCommons[i]) GpioSetPin(lcdCommons[i],0);
	}

	// set segments;
	for (i=0; i<16; i++) {		
		if (lcdSegments[i]) GpioSetPin(lcdSegments[i],hilo);
	}
			
}

u_int16 lcdmap[5*7] = {
	33,17,1,0,32,49,16,
	35,19,3,18,50,51,34,
	38,22,6,5,37,54,21,
	40,24,8,7,39,56,23,
	63,63,63,63,63,63,63
};


u_int16 lcdNumberGlyph[10] = {
	1+2+4+8+16+32, //0
	1+2, //1
	4+2+64+16+32, //2
	4+2+64+1+32, //3
	8+64+2+1, //4
	4+8+64+1+32, //5
	4+8+16+32+1+64, //6
	8+4+2+1, //7
	127, //8
	127-16
};

const u_int16 alphaGlyph[] = {
    32, 0, // char  
    33, 34, // char !
    34, 10, // char "
    35, 123, // char #
    36, 109, // char $
    37, 117, // char %
    38, 126, // char &
    39, 8, // char '
    40, 60, // char (
    41, 39, // char )
    42, 78, // char *
    43, 67, // char +
    44, 33, // char ,
    45, 1, // char -
    46, 0, // char .
    47, 82, // char /
    48, 63, // char 0
    49, 3, // char 1
    50, 118, // char 2
    51, 103, // char 3
    52, 75, // char 4
    53, 109, // char 5
    54, 125, // char 6
    55, 15, // char 7
    56, 127, // char 8
    57, 111, // char 9
    58, 36, // char :
    59, 33, // char ;
    60, 76, // char <
    61, 96, // char =
    62, 70, // char >
    63, 86, // char ?
    64, 119, // char @
    65, 95, // char A
    66, 121, // char B
    67, 112, // char C
    68, 115, // char D
    69, 124, // char E
    70, 92, // char F
    71, 61, // char G
    72, 89, // char H
    73, 1, // char I
    74, 51, // char J
    75, 60, // char K
    76, 56, // char L
    77, 65, // char M right side
    78, 81, // char N
    79, 113, // char O
    80, 94, // char P
    81, 79, // char Q
    82, 80, // char R
    83, 109, // char S
    84, 120, // char T
    85, 49, // char U
    86, 59, // char V
    // 87, 49, // char W 
    87, 33, // char W right side
    88, 91, // char X
    89, 107, // char Y
    90, 54, // char Z
    91, 60, // char [
    92, 73, // char \
    93, 39, // char ]
    94, 4, // char ^
    95, 32, // char 	_
    96, 4, // char 
    0,0
};





	
void LcdSetSegment(int segment, int state) {
	u_int16 k = lcdd[segment >> 4];
	k &= ~(1 << (segment & 0xf));
	if (state) k |= (1 << (segment & 0xf));
	lcdd[segment >> 4] = k;
}


void LcdPutNum(int position, int number) {
	int i;
	for (i=0; i<7; i++) {
		LcdSetSegment(lcdmap[position*7 + i], lcdNumberGlyph[number]&(1<<i));
	}
}


u_int16 lcdPos = 0;

void LcdPutChar(char ch) {
	int i;
	if (lcdPos > 3) return;
	ch = toupper(ch);
	
	if (ch == '.') {
		lcdPos = 99;
		return;
	}
	
	//if ((ch < 'A') || (ch > 'Z')) return;
	if (lcdPos == 99) return;


	if ((ch < 33) || (ch>96)) return;
	//printf("   '%c'",ch);
	ch -= 32;
	
	//printf("%d=%d(%c)  ",ch,alphaGlyph[ch*2+1],ch+32);

	for (i=0; i<7; i++) {
		LcdSetSegment(lcdmap[lcdPos*7 + i], alphaGlyph[ch*2+1]&(1<<i));
	}
	//LcdGenerate();
	lcdPos++;
	/*
	if (lcdPos == 4) {
		lcdPos = 0;
		DelayL(2000000);
		LcdClear();
	}
	*/
}


void LcdPutCharPos(int lcdPos, char ch) {
	int i;
	u_int16 glyph = 0;
	if (lcdPos > 3) return;
	//printf("%c",ch);
	for (i=0; alphaGlyph[i]; i+=2) {
		if (alphaGlyph[i] == ch) {
			glyph = alphaGlyph[i+1];
			break;
		}
	}
	for (i=0; i<7; i++) {
		LcdSetSegment(lcdmap[lcdPos*7 + i], glyph &(1<<i));
	}
}



void LcdPutString(char *s) {
	lcdPos = 0;
	LcdClear();
	while(*s) {
		LcdPutChar(*s++);
	}
}







void LcdClear() {
	lcdPos = 0;
	memset(lcdd,0,sizeof(lcdd));
}


#define TIMER1_INTERVAL 12288
void ConfigureTimer1() {
	static const VCODE vcode[] = {
		ClearBitsP3(TIMER_ENA, TIMER_ENA_T1),
		WriteP3(TIMER_T1H, (TIMER1_INTERVAL-1) >> 16),
		WriteP3(TIMER_T1L, (TIMER1_INTERVAL-1)),
		WriteP3(TIMER_T0CNTL, 1),
		WriteP3(TIMER_T0CNTH, 1),
		SetBitsP3(TIMER_ENA, TIMER_ENA_T1),
		SetBitsP3(INT_ENABLE0_HP, INTF_TIMER1),
		0
	};
	Xvm((void*)vcode);
	printf("T1 Conf\n");
}

// Timer 1 interrupt handler
#pragma interrupt y 0x2f
void MyTimer1InterruptHandler(void) {
	if (lcdPhase < 8) GpioSetPin(0x03,1); //SD driver fights to drive this. We fight back harder.
	LcdStep();
}

void InitLcdTimer() {
	GpioSetPin(0x20,0);
	GpioSetPin(0x21,0);
	GpioSetPin(0x22,0);
	GpioSetPin(0x23,0);
	LcdClear();
	LcdGenerate();
	LcdSetAllSignals(0);
	DelayL(100);
	ConfigureTimer1();
}


void LcdPutUnsigned4(u_int16 n){
	int i;
	char s[8];
	LcdClear();	
	sprintf(s,"%5d",n);
	//printf("Lcd:'%s' ",s);	
	for (i=0; i<4; i++) {
		char ch = s[i+1];
		if ((ch >= '0') && (ch <= '9')) {
			LcdPutNum(i,ch-'0');
		}
	}
	if (n>9999) LcdSetSegment(ICON_BATT1, 1);	
	if (n>19999) LcdSetSegment(ICON_BATT2, 1);
	if (n>29999) LcdSetSegment(ICON_BATT3, 1);
	LcdGenerate();
}



void StopLcdTimer() {
	PERIP(TIMER_ENA) &= ~TIMER_ENA_T1;
	PERIP(INT_ENABLE0_HP) &= ~INTF_TIMER1;
	PERIP(INT_ENABLE0_HP) &= ~INTF_TIMER1;
	LcdSetAllSignals(0);
}

