/// \file main.c VS1010D VSOS Executable main file
/// This is a starting point for developing VS1010D DLX libraries and executables

#include <vo_stdio.h>
#include <volink.h>     // Linker directives like DLLENTRY
#include <apploader.h>  // RunLibraryFunction etc
#include <vs1010dRom.h>
#include <vo_gpio.h>
#include <vs1010c.h>
#include <playerinfo.h>
#include <string.h>
#include <protocol.h>
#include <spitv.h>
#include <lcd.h>
#include <vo_fat.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <audio.h>
#include <voaudio.h>
#include <decodeAudio.h>
#include "fatdirectory.h"
#include "fatdircache.h"
#include "lcdui.h"
#include "lcdtimer.h"
#include "rtcread.h"
#include <appstate.h>
#include <fmcommon.c>

char songFileName[FAT_LFN_SIZE];


u_int32 t1, t2, t3;
int nextSong = 0;

void UiHandler();
void MP3PlayerCallback(AUDIO_DECODER *auDec, u_int16 samples) {
	//LcdPutMp3SongProgress();	
	
	UiHandler();
	if ((screenHoldSeconds == 0) && (app.nextMode != app.mode)) {
		//app.mode = app.nextMode;
		// Exit for switching to another app
		cancelSong:		
		lastReceivedCharUart0 = 0;
		auDec->cs.cancel = 1;		
	}
	

}

#include <codecmpg.h>
extern const struct CodecServices __y audioDecoderDefaultCodecServices;
auto AUDIO_DECODER *MyCreateMP3Decoder1010(register FILE *inFp, register FILE *outFp, register void (*callback)(AUDIO_DECODER *dec, u_int16 samples), register enum AuDecFormat auDecFormat) {
	AUDIO_DECODER *dec = &player.auDec;
  		
	memset(&player.auDec,0,sizeof(player.auDec));
	dec->cod = CodMpgCreateStatic();
	CodMpgSetTags(dec->cod, codMpgDisableCrc, codMpgEnableLayer2, codMpgLeaveOnError, 0);
  		
	memcpyYX(&dec->cs, &audioDecoderDefaultCodecServices, sizeof(audioDecoderDefaultCodecServices));
	
	dec->inFp = inFp;
	dec->outFp = outFp;
	dec->cs.fileSize = dec->cs.fileLeft = inFp->fileSize;
	dec->callback = (void*)callback;
	
	return dec;
}


void UpdateVolume(int volume) {
	int d;
	int h;	
	if (volume < 0) volume = 0;
	if (volume > 160) volume = 160;
	d = volume / 2;
	h = (volume & 1) * 5;
	screenHoldSeconds = 2;

	player.volume = volume;

	LcdClear();
	LcdSetSegment(ICON_DOT2,1);
	LcdSetSegment(ICON_DB,1);
	if (volume) {
		if (d > 9) {
			LcdSetSegment(ICON_MINUS,1);
		} else {
			LcdSetSegment(ICON_MINUS_2,1);
		}
	}
	
	if (d > 9) LcdPutNum(1,(d/10)%10);
	LcdPutNum(2,(d)%10);
	LcdPutNum(3,(h)%10);
	SetVolumeLR(player.volume, player.volume);
	LcdGenerate();
	app.masterVolume = player.volume;

}




void ButtonHandler(register u_int16 button) {
	static u_int16 timePressed;
	static u_int16 lastButton;
	
	if (button && button == lastButton) {
		timePressed++;
	} else {
		timePressed = 0;
	}
	lastButton = button;
	
	if (timePressed && (timePressed < 10)) return;
	
	if (button == 11) {
		player.pauseOn = 0;
		player.auDec.cs.cancel = 1;
		nextSong = dirc.nowPlaying + 1;
	}


	if (button == 5) {
		player.pauseOn = 0;
		player.auDec.cs.cancel = 1;
		nextSong = FirstSongOfNextAlbum(dirc.nowPlaying);
	}

	if (button == 6) {
		NextInput();
		if (app.inputSelect == INPUT_SD) {
			app.mode = app.inputSelect; //This dlx executable can handle this mode
		}
	}

	if (button == 10) {
		player.pauseOn = 0;
		player.auDec.cs.cancel = 1;
		if (dirc.nowPlaying > 1) nextSong = dirc.nowPlaying - 1;
	}
	if (button == 8) {
		UpdateVolume(player.volume + 1);
	}
	if (button == 9) {
		UpdateVolume(player.volume - 1);
	}


}


void ScreenPutTrack() {
	if (screenHoldSeconds) {
		//screenHoldSeconds--;
		return;
	}
	LcdPutUnsigned4(dirc.nowPlaying);
	LcdSetSegment(ICON_TRACK,1);
	if (app.mode == INPUT_USB_HOST) LcdSetSegment(ICON_USB, 1);
	LcdGenerate();
}

u_int16 fnHold;
u_int16 fnpos = 0;
void PutFileName() {
	LcdClear();

	LcdPutCharPos(0,songFileName[fnpos]);
	LcdPutCharPos(1,songFileName[fnpos+1]);
	LcdPutCharPos(2,songFileName[fnpos+2]);
	LcdPutCharPos(3,songFileName[fnpos+3]);
	LcdGenerate();
	fnpos++;
	if (!songFileName[fnpos]) fnpos = 0;
	if (fnpos > 80) fnpos = 0;
}


void UiHandler() {
	static u_int16 namestep = 0;	
	static u_int32 screenHoldSecondsTimeCount = 0;
	u_int32 t;
		
	
	if (lastReceivedCharUart0 == '2') { // Debug console -> go to mass storage to update firmware
		int i;
		StopLcdTimer();
		Disable();
		for (i=0; i < 64; i++) {  //Restore default memory layout
			PERIP(TLB_BASE+i) = 64*i;
		}
		RunLib("REBOOT","2"); // USB Mass storage, SD card, 1 bit mode, slow USB
	}
	
	if (((t = ReadTimeCount()) - screenHoldSecondsTimeCount) > 2000) {
		screenHoldSecondsTimeCount = t;
		if (screenHoldSeconds) {
			screenHoldSeconds--;
		}
	}

	if ((ReadTimeCount() - uiTimeCount) > 100) {
		uiTimeCount += 100;
		ButtonHandler(GetButtonPress());
		if ((!screenHoldSeconds) && (++namestep > 4)) {
			namestep = 0;
			if (fnHold) {
				fnHold--;
			} else {
				PutFileName();
			}
		}

		// update current playTimeSeconds to app settings, separate for sd card and usb flash playback			
		if (player.auDec.cs.playTimeSeconds) {
			if (app.mode == INPUT_USB_HOST) {
				app.usbSongPositionSeconds = player.auDec.cs.playTimeSeconds;
			} else {
				app.sdSongPositionSeconds = player.auDec.cs.playTimeSeconds;
			}
		}


	}
	
//	if ((ReadTimeCount() - uiTimeCount) < 10) {
//		RtcReadHandler();
//		if (lastTime != rtcShifter) {
//			lastTime = rtcShifter;
//			UpdateCurrentTimeOnly();
			//ScreenPutTime();
			//ScreenPutTrack();
//		}
//	}
}

 
void RestartStorage() {
	int i = S_ERROR;
	while (i == S_ERROR) {	
		u_int32 t1 = ReadTimeCount();
		
		if (app.shutdownRequested) return;			


		i = ioctl(VODEV('M'), IOCTL_RESTART, NULL);		
		if (!i) {
			LcdClear();
			LcdGenerate();
			screenHoldSeconds = 0;
			return;
		}
		printf("%s:%d\n",Identify(VODEV('M')),i);
		LcdClear();
		LcdPutCharPos(0,'N');
		LcdPutCharPos(1,'O');
		
		LcdPutCharPos(3,'F');
		if (app.mode == INPUT_USB_HOST) LcdSetSegment(ICON_USB, 1);
		LcdGenerate();
		screenHoldSeconds = 2;
		//LcdSetSegment(ICON_COLON,1);
		printf("Init %s M: %d\n",Identify(VODEV('M')),i);
		while ((ReadTimeCount() - t1) < 400) {
			UiHandler();
			if (app.nextMode != app.mode) return;
		}
	}

}

u_int16 onetime = 1;
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// MP3 player proper
void ToslinkInit(void);
u_int32 ReadShifter();
void WriteShifter(u_int32 data);
ioresult main (char *params) {
	int i;
	ToslinkInit();
	
//	read tracknumber from RTC RAM!	
//	app.sdTrackNumber = (u_int16)ReadShifter();
//	printf("read 1:%u\n ", app.sdTrackNumber);
//	write tracknumber to RTC RAM!
//			app.sdTrackNumber = 32;
//			WriteShifter(app.sdTrackNumber);
//			printf("write:%u\n ",app.sdTrackNumber);
//	read tracknumber from RTC RAM!	
	app.sdTrackNumber = (u_int16)ReadShifter();
	printf("app.sdTrackNumber from RAM:%u\n ", app.sdTrackNumber);

				
	ioctl(stdaudioout,IOCTL_RESTART,NULL);	
	SetPLL(3);

	app.mode = app.nextMode;
	uiTimeCount = ReadTimeCount();

	if (!VODEV('M')) { //If M: device is not set
		VODEV('M') = VODEV('D'); //then set it to SD card (device D:)
	}
	i = S_ERROR;
	InitLcdTimer();
	LcdSetAllSignals(0);

	RestartStorage();	
	//RunLib("DIR","M:");
	if (app.mode == INPUT_USB_HOST) {
		dirc.nowPlaying = app.usbTrackNumber;
		printf("usbt=%d ",dirc.nowPlaying);
	} else {
		dirc.nowPlaying = app.sdTrackNumber;
		printf("sdt=%d ",dirc.nowPlaying);
	}


	if (dirc.totalSongs == 0) {
		t1 = t2 = ReadTimeCount();
		BuildDirectoryCache();
		//printf("%ld milliseconds total indexing time.\n\n",(ReadTimeCount()-t1)/2L);
	}
	//printf("\r%u songs on disk.\n",dirc.totalSongs);
	
	while (app.nextMode == app.mode){
		FILE *f;
		LcdPutUnsigned4(dirc.nowPlaying);
		LcdSetSegment(ICON_TRACK,1);
		LcdGenerate();

		RestartStorage();
		
		if (app.shutdownRequested) break;			


		if (app.mode == INPUT_USB_HOST) {
			printf("usbt:%d ",dirc.nowPlaying);
			app.usbTrackNumber = dirc.nowPlaying;
		} else {
			printf("sdt:%d\n ",dirc.nowPlaying);
			app.sdTrackNumber = dirc.nowPlaying;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
//	write tracknumber to RTC RAM!
			WriteShifter(app.sdTrackNumber);
			printf("app.sdTrackNumber to RAM:%u\n ",app.sdTrackNumber);

		}

		ScreenPutTrack();
		f = OpenSongByNumber(dirc.nowPlaying);		
		screenHoldSeconds = 3;
		fnpos = 0;

		printf("play\n");
		
		if (f) {
			t1 = ReadTimeCount();
			auDec = MyCreateMP3Decoder1010(f, stdaudioout, MP3PlayerCallback, auDecFMp3);
			if (!auDec || !auDec->cod || !auDec->cod->Decode) {
				printf("Cannot decode");
			} else {
				char *eStr = NULL;
				uiTimeCount = ReadTimeCount();
				fnHold = 2;


				if (onetime) { //Once after (re)starting sd or usb player, seek to the last playTimeSeconds to continue song from previous position
					onetime = 0;
					if ((app.mode == INPUT_USB_HOST) && (app.usbTrackNumber = dirc.nowPlaying )){
						player.auDec.cs.goTo = app.usbSongPositionSeconds;
					}
					if ((app.mode != INPUT_USB_HOST) && (app.sdTrackNumber = dirc.nowPlaying )){
						player.auDec.cs.goTo = app.sdSongPositionSeconds;
					}
				}
	
				{
					u_int16 i = auDec->cod->Decode(auDec->cod, &auDec->cs, &eStr);
					printf(": %s(%d) ",eStr,i);
				}
				DeleteAudioDecoder1010();
				printf("after %ld milliseconds.",(ReadTimeCount()-t1)/2L);
			}
			printf("\n");
			fclose(f);
		}
		/*	
		if (app.mode == INPUT_USB_HOST) {
			app.usbSongPositionSeconds = 0;
		} else {
			app.sdSongPositionSeconds = 0;
		}
		*/

		if (nextSong) {
			dirc.nowPlaying = nextSong;
			nextSong = 0;
		} else {
			dirc.nowPlaying++;
		}
								
		if (dirc.nowPlaying > dirc.totalSongs) dirc.nowPlaying = 1;
	}
	
	app.mode = app.nextMode;
	StopLcdTimer();
	
	return S_OK;
}

void fini(void) {
	StopLcdTimer();
	
	PERIP(TIMER_ENA) &= ~TIMER_ENA_T0;
	PERIP(INT_ENABLE0_LP) &= ~INTF_TIMER0;
	PERIP(INT_ENABLE0_HP) &= ~INTF_TIMER0;

	PERIP(TIMER_ENA) &= ~TIMER_ENA_T1;
	PERIP(INT_ENABLE0_HP) &= ~INTF_TIMER1;
	PERIP(INT_ENABLE0_HP) &= ~INTF_TIMER1;
	LcdSetAllSignals(0);
}

