/// \file display.h Headers for oled/lcd display support functions
/// \author Panu-Kristian Poiksalo, VLSI Solution Oy

#ifndef DISPLAY_H
#define DISPLAY_H

#include "standalone.h"
#include "display.h"
#include "lcdisplay.h"

#define ScreenPutChar(c) ScreenOutputChar(c,ScreenPutData)
#define ScreenPutBigChar(c) ScreenOutputChar(c,ScreenPutDoubleData)

extern u_int16 screenX, screenY;
extern u_int16 screenNotificationString[40];
extern __y u_int16 screenNotificationTime;
extern __y u_int16 screenNotificationDigits;
extern __y u_int16 screenNotificationValue;


void ScreenNotification(const char *s, u_int16 value, u_int16 digits);
/// Put n binary octets to display controller
void ScreenPutsN(unsigned char *c, int n);
/// Write command c to display controller
void ScreenPutCommand(unsigned char c);
/// Set x,y pointer of display controller
void ScreenLocate(int x, int y);
/// Write image data octet to display controller
void ScreenPutData(u_int16 c);
// Write double height image data octet to display controller 
void ScreenPutDoubleData(u_int16 c);
/// Draw character c to screen using custom data output function
/// This function is responsible for character mapping.
void ScreenOutputChar(register u_int16 c, void (*OutputFunction)(u_int16 c) );
/// Screen Clear
void ScreenClear();
/// Put packed string to display (mainly for file name output)
void ScreenPutPacked(const char *d, u_int16 xEnd);
/// Draw a u_int16 value to the screen as decimal number
void ScreenPutUInt(register u_int16 value, register u_int16 length);
/// Draw a u_int16 value to the screen as decimal number in large numbers
void ScreenPutBigInt(register u_int16 value, register u_int16 length);
/// Draw a volume bar (triangle) to screen
void VolumeBar(u_int16 volume);
/// Draw a progress bar to the screen
/// Allow code optimization by making the progress bar always start at
/// the leftmost edge of screen
void ScreenPutBar(register u_int32 currentValue, u_int32 maxValue, u_int16 xMax);
// Write zero data (black pixels) to display until x-coordinate N
auto void ScreenZeros (u_int16 n);
// Start the display controller
void ScreenInit();



#endif

