/// \file standalone.c -- VS1053 Stand-Alone MP3 Player Source Code
/// \author Panu-Kristian Poiksalo, VLSI Solution Oy 2009-2010
/// Based on the VS1053 Stand-alone player source code by Pasi Ojala
/// Copyright (C) 2005-2009 VLSI Solution Oy Tampere Finland
// See readme.txt.

/*
  VS1053: I2S and higher reference
  VS1053B has extended GPIO registers:
  GPIDATA[8](SCLK),[9](XCS),[10](SI),[11](XDCS) and GPODATA[8](SO)
*/

#include "defines.h"

//#define NOCHECKFILETYPE 

#define SHUFFLE_PLAY /*shuffle instead of random +12/19 words*/
#define SKIP_ID3V2    /* skip ID3v2 tags */

#ifndef NO_WMA
//#define UI_PAUSE_BEFORE_PLAY //disable pause key function, wait for play
#endif

#define USE_LCD_DISPLAY
#define TIMEOUT /* timeout when decoding seems to be stuck */
//#define ENABLE_I2S /*practical for VS1053 only*/
#define ENABLE_HIGHREF /*VS1053 only -- use if AVDD >= 3.3V */
 
//#define SAVE_POSITION
//#define SAVE_VOLUME

//#define PLAYTIME 15
//#define SAVE_VOLUME /*66 words */
//#define SAVE_POSITION /*32 words*/
#if defined(SAVE_VOLUME) || defined(SAVE_POSITION)
#define SAVE_STUFF
#endif

#define ENABLE_UI

#include <string.h>
#include "standalone.h"
#include <stdio.h>
#include "mmcfs.h"
#include "display.h"


#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

s_int16 bassTarget = 0;
u_int16 bassDelay = 0;

u_int16 GetI6(void);
auto void DecodeTimeClear(void);
void PowerOff(void);

#define FAT_LFN_SIZE 100
u_int16 longFileName[FAT_LFN_SIZE];
s_int16 fileNamePos = -20;
u_int16 fileSizeW;


typedef enum {
  spi_pause=0,
  spi_file,
  spi_seek,
  spi_waitdata,
  spi_data,
  spi_zeros,
  spi_record,
  spi_waitima
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
#ifdef TIMEOUT
u_int16 timeOut;
#endif

struct SPIX { /* do not touch! ASM uses the structure */
  s_int32 size;
  u_int32 sector;
  u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
  __y struct FRAGMENT *fragment;
} spix;


u_int16 screenUpdate;


register __a1 short SpiDelay(void);

void Delay(u_int32 n){
  while (n--) {
    SpiDelay();
  }
}


void puthex(u_int16 d);
#if 0
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
  register i;
  for (i=0;i<4;i++) {
    putchar(hex[(d>>12)&15]);
    d <<= 4;
  }
}
void puthex2(u_int16 d) {
  register i;
  putchar(hex[(d>>4)&15]);
  putchar(hex[(d>>0)&15]);
}
#endif


#ifdef USE_LCD_DISPLAY

__y u_int16 screenPhase = 0;


void ScreenPutVolts() {
  static __y u_int32 voltMeas;
  static __y u_int16 lowVoltage = 0;  
  static __y u_int16 myPhase = 0;
  static __y u_int16 voltMeasCount = 0;
  
  
  myPhase++;

  USEX(DECIM_CONTROL) = DECIM_ENABLE | DECIM_FACTOR48K; // Decimator on, factor, enabled.

  if (16 == (myPhase & 0x1f)) {
    if (lowVoltage) {
      ScreenLocate(101,0);
      ScreenZeros(131);
    }
  }    


  voltMeas += USEX(DECIM_DATA_RIGHT);
  voltMeasCount++;
  
  if (0 == (myPhase & 0x1f)){  
    // Battery voltage

    if (voltMeasCount) {
      // calculate average of voltMeas
      voltMeas /= voltMeasCount;
    }

    // Linear convert into millivolts
    voltMeas += 8122;
    voltMeas *= 1000;
    voltMeas /= 2500;

    ScreenLocate(101,0);
    ScreenPutChar(voltMeas/1000+'0');
    ScreenPutChar('.');
    ScreenPutUInt(voltMeas%1000,3);
    ScreenPutData(0);
    ScreenPutChar('V');

    // Is battery voltage below 3.6 volts? Then set lowVoltage flag.
    lowVoltage = (voltMeas < 3600) ? 1 : 0;

    // Is battery voltage below 3.5 volts? Then set power off.
    if (voltMeas < 3500) {
      static u_int16 powerOffCount = 0;
      if (++powerOffCount > 10) {
	PowerOff();
      }
    }

    voltMeas = 0;    
    voltMeasCount = 0;
  }    
}


void ScreenUpdate() {

  // Song Number or Shuffle
  ScreenLocate(0,0);
  if (USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY) {
    ScreenPutPacked("\pShuffle",37);    
  } else {
    ScreenPutUInt(USEX(SCI_AICTRL0)+1,4);
  }
  ScreenZeros(37);
 

 
  if ((AudioBufFill() < 4000) && (spiState != spi_pause)) {
    return; //not enough CPU time to do screen update now.
  }
 
/*  ScreenLocate(0,1);
  ScreenPutUInt(USEX(SCI_HDAT1), 5);
  ScreenPutChar(' ');
  ScreenPutUInt((u_int16)stream_wr_pointer,5);
  ScreenPutChar(' ');
  ScreenPutUInt((u_int16)spiState,5);*/

  // Decode Time
  ScreenLocate(38,4);
  ScreenPutBigInt(USEX(SCI_DECODE_TIME)/60, 2);
  ScreenPutBigChar(':');
  ScreenPutBigInt(USEX(SCI_DECODE_TIME)%60, 2);
  

  ScreenPutVolts();
  screenPhase++;
  
  ScreenLocate(0,2);

  if (screenPhase & 1)

    if (screenNotificationTime) {
      screenNotificationTime--;
      ScreenPutPacked(screenNotificationString,0);
      //---scr---
      if (screenNotificationDigits) {
	ScreenPutUInt(screenNotificationValue, screenNotificationDigits);
      }
      ScreenZeros(132);
      

    } else {
      {
	u_int16 i,j = 0;
	
	fileNamePos++;
	if (fileNamePos > -20){
	  j = (fileNamePos<0) ? 0 : fileNamePos;
	  while (longFileName[j]){
	    if (screenX > 126) {
	      goto nexit;
	    }
	    ScreenPutChar(longFileName[j++]);
	  }
	  fileNamePos = -40;
	nexit:
	  ScreenZeros(131);
	}
      }
    }
  

#if 0
  // Show Sample Rate
  ScreenLocate(70,0);
  ScreenPutUInt(hwSampleRate/1000,2);
  ScreenPutData(0);
  ScreenPutPacked("\pkHz",3);
#endif
	
  //if (1 == (screenPhase & 0x1f)){
    
    // File name extension
    ScreenLocate(40,0);
    ScreenPutChar(fileName[4] >> 8);
    ScreenPutChar(fileName[4] & 0xff);
    ScreenPutChar(fileName[5] >> 8);
    ScreenZeros(69);
    
    // Pos
    ScreenLocate(0,7);
    ScreenPutBar(fatInfo.fileSize-spix.size*2, fatInfo.fileSize,130);      
    
    // Bitrate
    ScreenLocate(70,0);
    ScreenPutUInt(parametric_x.byteRate/125,3);
    ScreenPutData(0);
    ScreenPutChar('k');
    ScreenZeros(100);

    //  }        
  
}


#else
#define ScreenInit();
#define ScreenLocate(a,b);
#define ScreenPutPacked(a,b);
#define ScreenPutChar(a);
#define ScreenPutBigChar(a);
#define ScreenPutVolts();
#define ScreenClear();
#define ScreenPutCommand(a);

#endif
 

void PowerOff(void) {
  SpiSendClocks();
  USEX(SCI_STATUS) |= (1<<SCIST_APDOWN2); /*drivers off*/

  // Display off
  ScreenClear();
  ScreenLocate (40,4);
  ScreenPutPacked("\pPower Off.",128);
  SCREEN_SELECT;
  Delay(2000000);
  ScreenClear();
  ScreenPutCommand(0xae); //Display Off
  SCREEN_SELECT;
  Delay(1000000);
  ScreenPutCommand(0xad); //DC-DC off
  ScreenPutCommand(0x8a); //value

  USEX(SER_DREQ) = 0; /* Led / Aux key scan off*/

  // Wait until power button not pressed    
  SCREEN_SELECT;
  do {
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
  } while ((USEX(GPIO_IDATA) & (1 << 10)) != 0);
    

  USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
  while (1)
    ;
}





//void Disable();
//void Enable();
//void Reset();

#include <stdlib.h>


//#define USE_DEBUG

#if defined(USE_DEBUG) && !defined(UART_BUFFERED)
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC))
#else
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif


#define FIX_WAV_DECODETIME  /* SCI_UI: advance decodetime for short WAV's */

//#define HDATCHECK 800 /* 21words: give up after 800/2=400kB ~ 3-6sec */


#ifndef NOCHECKFILETYPE
#define CHECKFILETYPE /* vs1011 29 words? check file type according to short-filename suffix */
#endif/*NOCHECKFILETYPE*/
//#define KEEP_DRIVERS_OFF /* .. until MMC located */
#define WITH_SUBDIRECTORIES /* support subdirectories */


s_int16 hcShift;


#ifdef CHECKFILETYPE
#undef CHECKFILETYPE /*temporary undef*/
#define CHECKFILETYPE CheckFileType1053
#endif

#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif


//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

u_int16 gFileNum[2];


/*147 words*/
auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
		       __y struct FRAGMENT *nextFragment) {
  register __c1 int i = 0;
  register u_int32 currentSector = curFragment->start /*& 0x7fffffffUL */;
  static u_int16 lfnReset = 0;
  
#ifdef ENABLE_WATCHDOG
  USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
  goto first;
  
  while (1) {
    register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */
    
    /* We are now looking at FAT directory structure. */
    /* Is current file a regular file? */
    if (fn == 0)
      goto eod; /*end of directory*/
    if (fn != 0xe5) { /* Not deleted */
      
      
      
      // Long file name
      if ((GetByte(i+11/*Attr*/) & 0x3f) == 0x0f) {
	register u_int16 idx = ((fn & 0x3f) - 1) * 13;
	static const short places[] = {
	  1, 3, 5, 7, 9, 14, 16, 18, 20, 22, 24, 28, 30, -1
	};
	register const short *p = &places[0];
	while (*p >= 0 && idx < FAT_LFN_SIZE-1 /* space for NUL */) {
	  longFileName[idx++] =
	    GetByte(i + *p++);
	}
	/* last long entry ? */
	if ( fn & 0x40 /*LAST_LONG_ENTRY*/ ) {
	  if (idx >= FAT_LFN_SIZE)
	    idx = FAT_LFN_SIZE-1;
	  longFileName[idx] = '\0';
	}
	lfnReset = 0;
      } else {
	if (++lfnReset==2) {
	  longFileName[0] = 0;
	}
      }
      
      
#ifdef WITH_SUBDIRECTORIES
      /* Is it a subdirectory? */
      if ((GetByte(i+11/*Attr*/) & 0x10) && fn != '.') {
	/* TODO: do not call if FAT12! */
	if (HandleDir(nextFragment,
		      FragmentList(nextFragment,
				   ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
	  return -1;
	//puts("reread dir");
	memset(stream_buffer, 0, 512/2);
	ReadDiskSector(currentSector);/*reread sector*/
      } else /* it was a subdirectory */
#endif
	/* Attributes: NO directory, NO volume id, NO system, NO hidden */
	if ((GetByte(i+11/*Attr*/) & 0xde) == 0
#ifdef CHECKFILETYPE
	    /*
	      check supported files from name suffix:
	      vs1011/vs1002:  wav mp3
	      vs1003:         wav mp3 mid wma asf
	      vs1033: wav mp3 mid wma asf mp4 aac m4a
	    */
	    && CHECKFILETYPE(GetLong(i+8))
#endif
	    ) {
	  
	  /* It is a regular file. */
	  
	  gFileNum[1]++; /* total file count */
	  if (gFileNum[0]-- == 0) {
	  thisfile:
	    /* Put the 8.3-character filename to memory so the user
	       can read it. Two bytes per word, big-endian. */
	    {   /* 13 words */
	      register int j;
	      register u_int16 *p = stream_buffer + i/2;
	      register __y u_int16 *d = fileName;
	      for (j=0;j<6;j++) {
		*d++ = *p++;
	      }
	    }

	    /* ------------ FILE FOUND ------------- */
	    fatInfo.fileSize = GetLong(i+28); /*do first!*/
	    FragmentList(&fragments[0],
			 ((u_int32)GetWord(i+20)<<16) + GetWord(i+26));
	    return -1; /* File found, All OK return */
	  }
	} /* normal file */
    } /* 0xe5 / 0x00 deleted or unused entry */
    i = (i + 32) & 511;
    if (i == 0) {
      currentSector++;
      if (--(curFragment->size) == 0) {
	if ((s_int16)(curFragment->start>>16) < 0) {
        eod:
	  return gFileNum[1];
	}
	curFragment++;
	currentSector = curFragment->start /*& 0x7fffffffUL*/;
      }
    first:
#if 0 && defined(USE_DEBUG)
      puthex(currentSector);
      puts(" read");
#endif
      //puts("read dir");
memset(stream_buffer, 0, 512/2);
      ReadDiskSector(currentSector);
    }
  }
}


#ifdef HDATCHECK
u_int16 hdatCheck;
#endif

/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
u_int16 __x * __x stream_save;


__y struct {
  u_int16 Data; /* must be first entry! Used by c.s */
  u_int16 Cnt;
  u_int16 Strobe;
  u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
  u_int16 Key;
  s_int16 KeyTime; 
} ui;


#ifdef ENABLE_UI
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n);
s_int16 ApplAddrButRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
		       register __a0 s_int16 n);
#endif/*ENABLE_UI*/

s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
		    register __a0 s_int16 n);
s_int16 ApplAddrRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
		    register __a0 s_int16 n);

void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#ifdef SAVE_VOLUME
int volChanged = 0;
#endif

#ifdef SAVE_STUFF
  
auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);
  
#if 1
void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#else
void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value) {
  USEX(GPIO_ODATA) &= ~(SPI_xCS2|SPI_xCS);
  SpiSendReceiveSpi(SPI_WREN<<8, 8);
  USEX(GPIO_ODATA) |= SPI_xCS;
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  SpiSendReceiveSpi(SPI_WRITE<<8, 8);
  SpiSendReceiveSpi(addr/*volume=26*/, 16);
  SpiSendReceiveSpi(value, 16);
  USEX(GPIO_ODATA) |= SPI_xCS|SPI_xCS2;//SPI_CLK;
}
#endif
//SaveWord(26/*volume=26*/, USEX(SCI_VOL));
#endif


#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif

int rewmode = 0;
int rewcount = 0;


u_int16 pauseOn = 0;
u_int16 autoPowerOff;

void UserInterface(u_int16 highestAllowedKey) {

  register __c1 SPISTATE cachedSpiState = spiState;

#ifdef ENABLE_UI
  
  static u_int16 holdOn = 0;
  static u_int16 modeFlashTimer = 0;
  

  if (ui.Strobe) {
    // Real-Time Tick
  }  
    
  // PKP: -- UI: USER INTERFACE --
          
  if (ui.Strobe) { // PKP: It's time to check the buttons and do UI calculations
    
    screenUpdate = 1;
    ui.Data = ui.Strobe = 0;
    // PKP: -- UI: READ KEYS -- 
    
    USEX(SER_DREQ) = 0;		
    ui.Data = 
      USEX(GPIO_IDATA) & (1 << 10) ? SW1 :
      USEX(GPIO_IDATA) & (1 <<  8) ? SW2 :
      USEX(GPIO_IDATA) & (1 << 11) ? SW3 :
      USEX(GPIO_IDATA) & (1 <<  6) ? SW4 : 0;
    if (!ui.Data) {
      USEX(SER_DREQ) = 1;
      ui.Data = 
	USEX(GPIO_IDATA) & (1 << 10) ? SW5 :
	USEX(GPIO_IDATA) & (1 <<  8) ? SW6 :
	USEX(GPIO_IDATA) & (1 << 11) ? SW7 :
	USEX(GPIO_IDATA) & (1 <<  6) ? SW8 : 0;		  
    }
    if (ui.Data > highestAllowedKey) {
      ui.Data = 0;
    }   
    
    // PKP: -- UI: KEY FUNCTIONS --
    // User Interface functions
    
    parametric_x.playSpeed = 1;    
    
    if (holdOn) {
      if (ui.Data && (ui.Data != SW_HOLD)) { 
	ui.Data = 0; //In HOLD mode, block everything except hold
	ScreenNotification("\pKeylock on. Long-press [C].",0,0);
	screenNotificationTime = 10;
      }
    }
    
    if (ui.Data == ui.Key) { // PKP: Current Keystate is same as previous Keystate
      if (ui.KeyTime < 8) {
	ui.KeyTime++;
      } else // PKP: keystate has been consistent for 8 ticks - it's a long keypress //
	if (ui.Key
          #ifdef WAIT_FOR_BUTTON
	    && ui.KeyTime > 0
          #endif
	    ) {

	  /* long presses */
	  register u_int16 vol = USEX(SCI_VOL);
	  switch(ui.Key) {
	    

	  case SW_HOME: //Long press of HOME
	    PowerOff();
	    break;
	    
	  case SW_VOL_UP: //Long press of Volume Up
	    if (vol) {
	      vol -= 0x0101;
	      goto lsetvol;
	    }
	    break;
	  
	  case SW_VOL_DN: //Long press of Volume Down
	    //if (vol != (s_int16)0xffffU)
	    if ((vol&0xff) < 198){
	      vol += 0x0101;
	    }
	  lsetvol:
	    ScreenNotification("\pVolume (dB) -",((vol & 0xff)/2),2);
	    USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
#ifdef SAVE_VOLUME
	    volChanged = 1;
#endif/*SAVE_VOLUME*/
	    break;
	    
	  
	  case SW_HOLD: //Long press of Hold
	    if (ui.KeyTime == 8) {
	      ui.KeyTime = 9;
	      // toggle HOLD
	      #ifdef USE_LCD_DISPLAY
	        screenNotificationTime = 20;
		if (holdOn) {
		  holdOn = 0;
		  ScreenNotification("\pKeylock Off.",0,0);
		} else {
		  holdOn = 1;
		  ScreenNotification("\pKeylock On.",0,0);
		}
	     #else
		if (holdOn) {
		  holdOn = 0;
		} else {
		  holdOn = 1;
		}
             #endif
	    }	  

	    break;	    
	    
	  case SW_PREV:
	    if (rewmode == 0) {
	      rewmode = 1;
	    }
	    break;
	      	    
	  case SW_NEXT:  // Long press of NEXT -> FF
	    parametric_x.playSpeed = 6;
	    break;

	  default: ;
	  } //end switch-case	  	 	  
	}
      
    } else { // PKP: Keystate is not the same as previous keystate
      // PKP: Handle short keypresses

      if (ui.Data == 0
        #ifdef WAIT_FOR_BUTTON
	  && ui.KeyTime > 0
        #endif
	  ) {
	// PKP: Handle end of short keypress (current keystate ui.Data=0)
	/* short presses */
	if (ui.KeyTime < 8) {
	  /* was pressed < 0.5 seconds */
	  register u_int16 vol = USEX(SCI_VOL);

	  switch (ui.Key) {
	  
	  case SW_REC:	    
	    screenNotificationTime = 10;
	    {
	      if (bassTarget == LOUDNESS_MIN_VALUE) {
		bassTarget = LOUDNESS_MEDIUM_VALUE;
		ScreenNotification("\pLoudness Normal.",0,0);
	      } else if (bassTarget == LOUDNESS_MEDIUM_VALUE) {
		bassTarget = LOUDNESS_HIGH_VALUE;
		ScreenNotification("\pLoudness High.",0,0);
	      } else {
		bassTarget = LOUDNESS_MIN_VALUE;
		ScreenNotification("\pLoudness Min.",0,0);
	      }
	    }
	    break;


	  case SW_VOL_UP:
	    if (vol>0x0403) {
	      vol -= 0x0404;
	      goto lsetvols;
	    }
	    break;
	  
	  case SW_VOL_DN:
	    if ((vol&0xff) < 194){
	      vol += 0x0404;
	      //goto lsetvols;
	    }
	    	    
	  lsetvols:
	    ScreenNotification("\pVolume (dB) -",((vol & 0xff)/2),2);
	    USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
#ifdef SAVE_VOLUME
	    volChanged = 1;
#endif/*SAVE_VOLUME*/
	    break;
	  
	  case SW_MODE: //Short press of Mode
	    modeFlashTimer = 16;
	    
	    // Toggle shuffle play
	    USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
	    break;
	  
	  case SW_NEXT: //intentional fall-through
	  case SW_PREV:
	    {
	      register u_int16 song = USEX(SCI_AICTRL0);
	      if (ui.Key == SW_NEXT) {
		song++; /* next file */
	      } else
               #ifndef UI_PAUSE_BEFORE_PLAY
		if (USEX(SCI_DECODE_TIME) < 5) //Song played for less than 5 seconds
               #endif
		  {
		    if (song == 0) {
		      song = USEX(SCI_AICTRL1)-1;
		    } else { /* previous file */
		      song--;
		    }
		    /* else the same file */
		}
	      USEX(SCI_AICTRL0) = song;
	      
	      /* especially needed for MIDI!! */
              #if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
	        USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
              #endif
	      cachedSpiState = spi_zeros;
              #if defined(ENABLE_LAYER12)
	        /* gain=0 for partial last frame (bugs) */
	        volume[0] = volume[1] = 0;
              #endif		       			
	    }
	    break;
	    
	  
	  case SW_PLAY:
	    /* play pause/continue -- not very good for midi.. */
	    pauseOn ^= 1;

	    break;
	  default: ;
	  } //end of switch-case

	}//(ui.keyTime was < 8) (Short keypress)

      }
      ui.Key = ui.Data;
      ui.KeyTime = 0;
    }
   
#if defined(SAVE_VOLUME)
    /*Save volume when volume button is released. */
    if (ui.Data == 0 &&
	volChanged &&
	cachedSpiState == spi_seek) {
      volChanged = 0;
      SaveWord(26/*volume=26*/, USEX(SCI_VOL));
    }
#endif/*SAVE_VOLUME*/
    
  }   
  spiState = cachedSpiState;

#endif /*ENABLE_UI*/
}

/*
  Called before HALT
*/
void UserHook(void) {
  static SPISTATE cachedSpiState;





  // Loudness
  if (ui.Strobe) {
    if (bassDelay) {
      bassDelay--;
    } else {      
      register s_int16 bassLevel = USEX(SCI_BASS) & 0xf0f0;
      bassLevel |= bassTarget & 0x0f0f;
      
      if ((s_int16)(bassTarget & 0xf000) > (s_int16)(bassLevel & 0xf000)) {
	bassLevel += 0x1000;
      } else if ((s_int16)(bassTarget & 0xf000) < (s_int16)(bassLevel & 0xf000)) {
	bassLevel -= 0x1000;
      }
      if ((bassTarget & 0x00f0) > (bassLevel & 0x00f0)) {
	bassLevel += 0x10;
      } else if ((bassTarget & 0x00f0) < (bassLevel & 0x00f0)) {
	bassLevel -= 0x10;
      }
      USEX(SCI_BASS) = bassLevel;	  
    }    
  }



#ifdef ENABLE_UI
  UserInterface(SW8);
#endif /*ENABLE_UI*/

  cachedSpiState = spiState;  

  
#ifdef PLAYTIME
  if (USEX(SCI_DECODE_TIME) > PLAYTIME && cachedSpiState != spi_zeros) {
    USEX(SCI_AICTRL0)++;
#if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
    USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
#endif
    cachedSpiState = spi_zeros;
  }
#endif/*PLAYTIME*/
  
  if (cachedSpiState == spi_file) {    

    if (0==longFileName[0]) {
      u_int16 i;
      u_int16 *p = &longFileName[0];
      for (i=0; i<4; i++){
	*p++ = fileName[i] >> 8;
	*p++ = fileName[i] & 0xff;
      }
      *p=0;
    }      

    /* extra byte read if odd file size */
    spix.size = (fatInfo.fileSize+1)/2; /* in words! */
    //	fileSizeK = spix.size/512;
    cachedSpiState = spi_seek;
  setfrag:
    spix.fragment = &fragments[0];
    spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
    spix.fragSize = fragments[0].size; /* *fatInfo.sectorsPerCluster*/
  }
  
  
  


  if (cachedSpiState == spi_seek || cachedSpiState == spi_pause) {
    // MMC/SD card sector read is over, SPI bus can be used for other operations
    
    if (pauseOn) {

      if (cachedSpiState != spi_pause) {
	oldSpiState = cachedSpiState;
	cachedSpiState = spi_pause;
	stream_save = stream_wr_pointer;
	stream_wr_pointer = stream_rd_pointer+1;
	autoPowerOff = 4515; //4500;
	
	clockX = 2; //1.0x
	SetRate(hwSampleRate);
	//USEX(SCI_STATUS) |= (1<<SCIST_APDOWN2);
      } 
    } else {
      if (cachedSpiState == spi_pause) {
	cachedSpiState = oldSpiState;
	/* restore 'stolen' data */
	stream_wr_pointer = stream_save;
	screenNotificationTime = 0;
	SetClockFromSpi();    
      }

    }
    
    
    SpiSendClocks(); /*Raise MMC/SD CS (and send some clocks) */
    
    USEX(GPIO_ODATA) |= SPI_xCS2;
    
#ifdef USE_LCD_DISPLAY
    if (screenUpdate) {

      if (pauseOn) {
	if (!--autoPowerOff) {
	  PowerOff();
	}
	ScreenNotification("\pPaused. Power off in ",autoPowerOff/16,3);
      }
	
      ScreenUpdate();
      screenUpdate = 0;
    }
#endif     
    SpiSendClocks();
    
 
    
  }
   
  if (cachedSpiState == spi_seek) {
    if (rewmode == 1) {
      s_int32 skip = spix.fragment->size - spix.fragSize;
      if (skip > REWIND_STEP_SIZE)
	skip = REWIND_STEP_SIZE;
      spix.sector -= skip;
      spix.fragSize += skip;
      spix.size += skip*512/2;
      
      rewcount = REWIND_STEP_SIZE / 8;
      rewmode = 2;
    } else if (rewmode == 2) {
      if (--rewcount == 0)
	rewmode = 0;
    }
    
   /* In case we are just trying to find something to play,
       also read keys once in a while.. */
    ui.Cnt += 64;
    if (ui.Cnt > hwSampleRate/16) {
      ui.Cnt -= hwSampleRate/16;
      ui.Timer++;
      ui.Strobe = 1;
    }

    //puts("ReadBlock");
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
	       (spix.sector & 0x7fffffffUL) << hcShift);
    cachedSpiState = spi_waitdata;
    spiCnt = 0;
    NextSector();
#ifdef HDATCHECK
    if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
      /* if file not detected at HDATCHECK/2 kB, skip the rest */
      goto end_of_file;
    }
#endif
  }
  
#ifndef VOICE_ACTIVATED
  /*TODO: handle spi_waitdata and spi_data
    in timer interrupt to increase throughput for FLAC */
  if (cachedSpiState == spi_waitdata) {
    if (SpiSendReceiveMmc(0xff00,8) == 0xfe) {
      cachedSpiState = spi_data;
      spiCnt = 0;
#ifdef TIMEOUT
      timeOut = 0;
#endif
    } else {
      if ((s_int16)++spiCnt < 0) {
	goto goreset; //MyReset(); /* timeout */
      }
    }
  } else if (cachedSpiState == spi_data) {
    while (spix.size > 0) {
      register int i, j = 32;
      register u_int16 *wr;

	{
	  if (StreamDiff() >= STREAM_BUFFER_SZ-40)
	    break;

	  wr = (void *)stream_wr_pointer;
	  if (j > spix.size)
	    j = spix.size;
	  for (i=0;i<j;i++) {
	    *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
	    if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
	      wr = stream_buffer;
	  }
	}
      stream_wr_pointer = (void *)wr;
      spix.size -= j;//32;
      spiCnt += 32;
      if (spiCnt >= 256) {
	/* 256 words read, get next block in file */
	cachedSpiState = spi_seek;
	SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
	break;
      }
#ifdef TIMEOUT
      timeOut = 0;
#endif
      /* watchdog clearing here doesn't work with midi */
    }
    
#ifdef TIMEOUT
    if ((s_int16)++timeOut == -1) {
      spix.size = 0;
    }
#endif
    
    if (spix.size < 0x20000L) {
      /* Workaround for WMA resync after stream ended! */
      parametric_x.resync = 0;
    }
    
    
#ifdef ENABLE_WATCHDOG
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    if (spix.size <= 0) {
    end_of_file:
      SpiSendClocks();
      

	{
	  {
	    register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
	    if (m == CTRL3_LOOP_SONG) {
	      /* restart the same file immediately */
	      cachedSpiState = spi_file;
	    } else if (m == CTRL3_PAUSE_AFTER_PLAY) {
	      oldSpiState = spi_zeros;
	      cachedSpiState = spi_pause;
	    } else {
	      USEX(SCI_AICTRL0) += 1; /* next file */
	      cachedSpiState = spi_zeros;
	      if (USEX(SCI_HDAT1) == (('f'<<8) | 'L'))
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
	    }
	  }
	}
    }

#ifdef NO_WMA /* reject WMA files */
    if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
      USEX(SCI_AICTRL0) += 1; /* next file */
      goto goreset;//MyReset();
    }
#endif
#if defined(NO_AAC) && !defined(SCI_UI) && !defined(RECORDER)
    {
      register u_int16 t = USEX(SCI_HDAT1);
      //0x4154(AT) 0x4144(AD) 0x4d34(M4)
      if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
	USEX(SCI_AICTRL0) += 1; /* next file */
	goto goreset;//MyReset();
      }
    }
#endif
  } else if (cachedSpiState == spi_zeros) {
    /* Workaround for WMA resync after stream ended! */
    parametric_x.resync = 0;
    
    if (StreamDiff() < STREAM_BUFFER_SZ-40) {
      register int i, z = parametric_x.endFillByte;
      register u_int16 *wr = (void *)stream_wr_pointer;
      for (i=0;i<32;i++) {
	*wr = z;
	if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
	  wr = stream_buffer;
	} else {
	  wr++;
	}
	++spiCnt;
	stream_wr_pointer = (void *)wr;
      }
      /* watchdog clearing here doesn't work with midi */
#ifdef TIMEOUT
      timeOut = 0;
#endif
    }
#ifdef TIMEOUT
    if ((s_int16)++timeOut < 0) {
      //putch(USEX(SCI_HDAT1)>>8);
      //putch(USEX(SCI_HDAT1));
      goto goreset;
    }
#endif
    
    if (USEX(SCI_HDAT1) == 0 ||
#if 1
	USEX(SCI_HDAT1) == 0x574d ||
#endif
	spiCnt >= 256+2048) { /* was left to max 256 by spi_data */
      /* 2048 words instead of bytes */
    goreset:
      /* Discard any unread data, including CRC, before starting again */
      {
	register int i;
	for (i=0; i<256+2; i++)
	  SpiSendReceiveMmc(-1/*0xffff*/,16);
      }
      SpiSendClocks();

#ifdef UART_UI
      putch('e');//USEX(UART_DATA) = 'e';
#endif
      USEX(DAC_VOL) = 0xaaaa;
      Delay(10000);
      MyReset();
    }
#ifdef ENABLE_WATCHDOG
    goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
  } else {
#ifdef ENABLE_WATCHDOG
  wdogack:
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
  }
#endif /* VOICE_ACTIVATED*/
    
#ifndef NO_DREQ_LED
  /* 15 words */
  {
    register int t = 0;
    
    if ((ui.Timer & 0x18) == 0) {
      /* DREQ indicates random play status */
#ifdef NO_RANDOM
      t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#else
      t = USEX(SCI_AICTRL3);  /* lowest bit is global random play */
#ifdef LOUDNESS_ALWAYS_ON_OFF
      /* indicate random play */
#else /*LOUDNESS_ALWAYS_ON_OFF*/
                  
      if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
	  ==
#else
	  !=
#endif
	  spi_pause)
	t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#endif /*else LOUDNESS_ALWAYS_ON_OFF*/
#endif /*else NO_RANDOM*/
    }
    USEX(SER_DREQ) = t;
  }
#endif/*!NO_DREQ_LED*/
    
  spiState = cachedSpiState;

}

#ifdef VS1053
void NewSinTest(void);
auto int PatchFrame(void);
#endif


s_int16 ApplAddrButRef (register __i0 s_int16 **d, register __a1 s_int16 mode,
				        register __a0 s_int16 n);


auto void MyMain(register s_int16 __a1 fullReset) {

  static u_int16 r;
  static u_int16 displayState = 0;

  //USEX(SCI_BASS) |= 0x031f;

  fileNamePos = -20; //File name position on screen
  longFileName[0] = 0;


  USEX(DECIM_CONTROL) = DECIM_ENABLE | DECIM_FACTOR48K; // Decimator on, factor, enabled.
  USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power (not drivers) */
  USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
  USEX(SCI_DECODE_TIME) = 0;


  applAddr = ApplAddrRef;
  USEX(SCI_STATUS) |= (1 << SCIST_REFERENCE_SEL);

 
  
  SpiSendClocks();
  if (0 == displayState) {
    ScreenInit();
    ScreenLocate(16,2);
    ScreenPutPacked("\pVLSI Solution VS1053",130);
    ScreenLocate (17,4);
    ScreenPutPacked("\pUSB HiFi Player v1.0",130);

    displayState = 1;
  }

  USEX(SER_DREQ) = 0; /* Led / Aux key scan off*/
  do { //Wait until power key is NOT pressed;
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
  } while ((USEX(GPIO_IDATA) & (1 << 10)) != 0);
      
      
  USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);


  USEX(SCI_MODE) = SciModeOr() | (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV));
  USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_xCS2|(1<<LCD_GPIO_PIN_A0)|(1<<LCD_GPIO_PIN_CS); /*use same= -1 word*/
  USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2|(1<<LCD_GPIO_PIN_A0)|(1<<LCD_GPIO_PIN_CS);
    
  InitHardware(); /* no longer clears memory in vs1053 */


#ifndef USE_DEBUG
  /* Allow reflashing without MMC/SD inserted. */
  USEX(INT_ENABLE) = (1<<INT_EN_RX);
#endif

  InitLayer3();
  InitCommon();
  InitDecode();

    

  /* If not powerup reset, re-enable analog immediately. */
  if (!fullReset) {
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
  }


  applAddr = ApplAddrButRef;

   
  /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
  USEX(INT_GLOB_ENA) = 0;
  USEX(INT_GLOB_ENA) = 0;

  InitReadFrame();

  DecodeTimeClear();
#ifdef USE_LCD_DISPLAY
      SpiSendClocks();
      SpiSendClocks();
#endif


  /* wake up MMC! -- If this takes too long, watchdog will bite. */

#ifdef SAVE_POSITION
  if (USEX(SCI_AICTRL0)) /* 9 words */
    SaveWord(28/*AICTRL0*/, 0); //3 words
#endif
    

  //while (1)
  {
    static u_int16 retries = 5;
    register s_int16 i, cmd;
    s_int32 d;
      
    // Start the MMC/SD card
  tryagain: 

    //USEX(FREQCTLH) = 1; //hack:low clock
    /*
      powerdown:
      clockX = 2; //1.0x
      SetRate(hwSampleRate);
      USEX(SCI_STATUS) |= (1<<SCISTF_ANA_PDOWN2);

      back:
      SetClockFromSpi();
     */


    ui.Strobe = 1;
    UserInterface(SW1);
    ScreenPutVolts();


    /* while in USB */
    if (USEX(GPIO_IDATA) & (1 << GPIO_PIN_USB_DETECT)){    
      //static u_int16 freqCtlSave;
      //freqCtlSave = USEX(FREQCTLH);
      //USEX(FREQCTLH) &= 0xfc0f; //switch off PLL

      ScreenInit();	
      ScreenLocate(16,2);
      ScreenPutPacked("\pVLSI Solution VS1053",130);
      ScreenLocate(50,4);
      ScreenPutBigChar('U');
      ScreenPutBigChar('S');
      ScreenPutBigChar('B');      
      
      while (USEX(GPIO_IDATA) & (1 << 9)){	
	ScreenPutVolts();
	ScreenLocate(62,7);
	if (USEX(GPIO_IDATA) & (1 << GPIO_PIN_USB_ACTIVITY)) {
	  //ScreenPutChar(SYMBOL_DISKDRIVE);
	  ScreenPutChar(SYMBOL_USB);
	} else {
	  ScreenPutPacked("\p",100);
	}
	Delay(80000);
      }
      ScreenInit();
    }
    

    // MMC reset
    USEX(GPIO_DDR) |= (1 << GPIO_PIN_SDCARD_OFF)/*SO, output*/;
    USEX(SER_DREQ) = 0;
    if (++retries>=25) {
      u_int16 gpio0_save = USEX(GPIO_ODATA);
      retries = 0;
      USEX(GPIO_ODATA) = 1 + (1 << GPIO_PIN_SDCARD_OFF); //all 0 except eecs and GPIO OFF
      Delay(100000); //keep SD card switched off for a while
      USEX(GPIO_ODATA) = gpio0_save;
    }
    USEX(GPIO_ODATA) &= ~(1 << GPIO_PIN_SDCARD_OFF); //SD ON
    Delay(10000);     
    for (i=0; i<16; i++) {
      SpiSendClocks();
    }
      
    /* MMC Init, command 0x40 should return 0x01 if all is ok. */
    i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
    for (cmd=0; cmd<512; cmd++) {
      SpiSendReceiveMmc(-1,16);
    }
    

    SpiSendClocks();
    if (i != 255) {
      ScreenLocate (15,4);
      ScreenPutPacked("\p  Card not ready (",1);
      ScreenPutUInt(i,2);
      ScreenPutPacked("\p)",130);
    } else {
      ScreenLocate(16,2);
      ScreenPutPacked("\pVLSI Solution VS1053",130);
      ScreenLocate(15,4);
      ScreenPutPacked("\pPlease insert SD Card",130);
    }
    
    if (i != 1)
      goto tryagain;//continue; /* No valid idle response */
      
    cmd = MMC_SEND_OP_COND|0x40;

    if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
	== 1) {
      /* MMC answers: 0x05 SD answers: 0x01 */
      /* Should we read the whole R7 response? */
      //SpiSendReceiveMmc(-1, 32);
      cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
    }

	
    /* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
    //i = 0; /* i is 1 when entered .. but does not make the code shorter*/
    while (1) {
      register int c;

      if (cmd == (0x40|41)) {
	MmcCommand(0x40|55/*CMD55*/,0);
	c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0x40000000UL);
      } else

	c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0);
      if (c == 0)
	break;

      if (++i >= 30000/*< 0*/ || c != 1) {
	goto tryagain; /* Not able to power up mmc */
      }
    }
	

    hcShift = 9;
    i = 0x0246;
    /* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
    /* Support HC for SD, but not for MMC! */
    if (cmd == (0x40|41) &&
	MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
      if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
	/* OCR[30]:CCS - card capacity select */
	/* HighCapacity! */
	/* ldc 9,D0 (0x0246) <-> ldc 0,D0 (0x0006) */
	i = 6;
	hcShift = 0;
      }
    }
    WriteToProgramRam((u_int16)ReadDiskSector+2, 0, i);

        

    /* Set Block Size of 512 bytes -- default for at least HC */
    /* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
    /* Must check return value! (MicroSD restart!) */
    i = MmcCommand(MMC_SET_BLOCKLEN|0x40, 512);
    if (i != 0) {      
      goto tryagain;
    }
    /* All OK return */

  }


  // Init File System (FAT)
  SpiSendClocks();
  ScreenLocate(16,2);
  ScreenPutPacked("\p   Scan File System",130);

  memset(stream_buffer, 0, 512/2);
  r = InitFileSystem();

  /* Detects cards without partition table */
  if (r == 0) {
    {
      ScreenClear();
      ScreenLocate(15,4);
      ScreenPutPacked("\p Searching for files",130);
      
      /* determine number of playable files */
      USEX(SCI_AICTRL1) = OpenFile(0xffffU);
    }
	

    if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
      /*shuffle 23 words*/
      USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
    }


    ScreenLocate(0,4);
    ScreenZeros(132);
    ScreenLocate(16,2);
    ScreenPutPacked("\p      Opening File",130);
    displayState = 0;


    // VS1053: Must set bass enhancer slowly to avoid noise
    // when volume is high
    if (USEX(SCI_BASS)){
      USEX(SCI_BASS) = LOUDNESS_MIN_VALUE;
    }
    bassDelay = 5;


    if (OpenFile(USEX(SCI_AICTRL0)) < 0) {
	      pauseOn = 0;
      /* check for ID3v2 */
      /* An ID3v2 tag can be detected with the following pattern:
	 $49 44 33 yy yy xx zz zz zz zz */

      ReadDiskSector(fragments[0].start);
      if ((GetLong(0) & 0xffffffL) == 0x334449) {
	/* skip in 16kB resolution, assume tags are smaller than 2MB */
	/* 0.5kB resolution now */
	register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
	/* guard against small fragments */
	if (skip >= fragments[0].size)
	  skip = fragments[0].size-1;
	fragments[0].start += skip;
	fragments[0].size  -= skip;
	fatInfo.fileSize -= (u_int32)skip * 512;
      }


      if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
	  CTRL3_PAUSE_BEFORE_PLAY) {
	USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
	oldSpiState = spi_file;
	spiState = spi_pause;
	stream_save = stream_wr_pointer;
      } else {
	spiState = spi_file;
      }
      hdatCheck = 0;

    } else {
    gofirst:
      USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
      MyReset();
    }
  } else {
    /* Initialize again */
    goto gofirst;//MyReset();
  }


      
#ifdef SAVE_POSITION
  SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
#endif

  USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
  USEX(SCI_DECODE_TIME) = 0;

  parametric_x.resync = 32767;

  stream_rd_pointer = stream_wr_pointer = stream_buffer;
  stream_rd_odd = 0;
      

  //Main loop for MP3. Other codecs DO break out from this loop
  while (1) {
    if (
	!ReadFrame()
	) {
      play_frame();
    } else {
      InitReadFrame();
    }
  }
}